//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include <algorithm>
#include <sstream>
#include <stdexcept>
#include <ctime>
#include <cstring>

#include "lfc1/datetime/cdate.hpp"
#include "lfc1/udtsup/input.hpp"
#include "lfc1/udtsup/output.hpp"

using namespace std;
using namespace lfc1::udtsup;

namespace lfc1 { namespace datetime {

///	@brief This function obtains the allocated iword/pword index.
///
///	@return	The allocated iword/pword index.
///
int CDate::smvGetIndex() noexcept
{
	call_once(smvFlag, smvAllocateIndex);
	return smvIndex;
}

///	@brief This function determines whether the given year is a leap year or not.
///
///	@param[in]	avYear	The year.
///
///	@retval	true	The year is a leap year.
///	@retval	false	The year is not a leap year.
///
bool CDate::smvIsLeapYear(int avYear) noexcept
{
	// Check if the year is a leap year.
	if (((avYear % 4) == 0 && (avYear % 100) != 0) || (avYear % 400) == 0)
	{
		return true;
	}
	else
	{
		return false;
	}
}

///	@brief This function obtains the maximum number of days for a month of a given year.
///
///	@param[in]	avYear	The year.
///	@param[in]	avMonth	The month.
///
///	@return	The maximum number of days for the given year and month.
///
int CDate::smvGetMaxDay(int avYear, int avMonth) noexcept
{
	// Obtain maximum number of days.
	static const int vMaxDays[] {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	auto vMaxDay = vMaxDays[avMonth];
	if (avMonth == 2 && smvIsLeapYear(avYear))
	{
		// Adjust the number of days in February for leap years.
		++vMaxDay;
	}
	return vMaxDay;
}

///	@brief This function validates a date to be in the supported date range.
///
///	@param[in]	avYear		The year of a date.  (0 - 9999)
///	@param[in]	avMonth		The month of a date.  (1 - 12)
///	@param[in]	avDay		The day of a date.  (1 - 31)
///
///	@exception	std::out_of_range	Indicates either the year, month or day are not within valid ranges.
///
void CDate::smvValidateDate(int avYear, int avMonth, int avDay)
{
	// Check the range for the year.
	if (avYear < MIN_YEAR || avYear > MAX_YEAR)
	{
		ostringstream vStream;
		vStream << "Year 'avYear' must be in the range 0 to 9999 at (" << __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}

	// Check the range for the month.
	if (avMonth < MIN_MONTH || avMonth > MAX_MONTH)
	{
		ostringstream vStream;
		vStream << "Month 'avMonth' must be in the range 1 to 12 at (" << __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}

	// Check the range for the day.
	auto vMaxDay = smvGetMaxDay(avYear, avMonth);
	if (avDay < MIN_DAY || avDay > vMaxDay)
	{
		ostringstream vStream;
		vStream
			<< "Day 'avDay' must be in the range 1 to " << vMaxDay << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
}

///	@brief This function creates a CDate object based on the current system date.
///
CDate::CDate() noexcept
:	mvDateNo {0}
{
	mvSetDate();
}

///	@brief This function creates a CDate object using the given Julian day number.
///
///	@param[in]	avDateNo	The Julian day number to assign to the CDate object.
///
///	@exception	std::out_of_range	Indicates that the given Julian day number is not within the valid range.
///
CDate::CDate(int avDateNo)
:	mvDateNo {avDateNo}
{
	if (avDateNo < MIN_DATENO || avDateNo > MAX_DATENO)
	{
		ostringstream vStream;
		vStream
			<< "Julian day number 'avDateNo' must be within the range " << MIN_DATENO << " to " << MAX_DATENO
			<< " at (" << __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
}

///	@brief This function creates a CDate object using the given year, month and day.
///
///	@param[in]	avYear	The given year.
///	@param[in]	avMonth	The given month.
///	@param[in]	avDay	The given day.
///
CDate::CDate(int avYear, int avMonth, int avDay)
:	mvDateNo {0}
{
	mvSetDate(avYear, avMonth, avDay);
}

///	@brief This function copy constructs a CDate object.
///
///	@param[in]	arRHS	The CDate object to be copied.
///
CDate::CDate(const CDate& arRHS) noexcept
:	mvDateNo {arRHS.mvDateNo}
{
}

///	@brief This function move constructs a CDate object.
///
///	@param[in]	arRHS	The CDate object to be moved.
///
CDate::CDate(CDate&& arRHS) noexcept
:	mvDateNo {arRHS.mvDateNo}
{
}

///	@brief This function destroys a CDate object.
///
CDate::~CDate() noexcept
{
}

///	@brief This function assigns the given Julian day number to the CDate object.
///
///	@param[in]	avDateNo	The Julian day number to assign to the CDate object.
///
///	@return	The CDate object assigned to.
///
CDate& CDate::operator=(int avDateNo)
{
	CDate vCopy {avDateNo};
	swap(vCopy);
	return *this;
}

///	@brief This function assigns a CDate object to another CDate object.
///
///	@param[in]	arRHS	The CDate object to assign to another CDate object.
///
///	@return	The CDate object assigned to.
///
CDate& CDate::operator=(const CDate& arRHS) noexcept
{
	mvDateNo = arRHS.mvDateNo;
	return *this;
}

///	@brief This function moves a CDate object to another CDate object.
///
///	@param[in]	arRHS	The CDate object to move to another CDate object.
///
///	@return	The CDate object assigned to.
///
CDate& CDate::operator=(CDate&& arRHS) noexcept
{
	mvDateNo = arRHS.mvDateNo;
	return *this;
}

///	@brief This function obtains the equivalent Julian day number of a CDate object.
///
///	@return The equivalent Julian day number of a CDate object.
///
CDate::operator int () const noexcept
{
	return mvDateNo;
}

///	@brief This function sets a CDate object based on the current system date.
///
void CDate::mvSetDate() noexcept
{
	// Obtain the current system time.
	auto vSystemTime = time(0);

	// Convert the current system time to a structure.
	tm vTm;
	memset(&vTm, 0, sizeof(vTm));
#if defined(__LINUX__) || defined(__UNIX__)
	localtime_r(&vSystemTime, &vTm);
#endif
#if defined(__WIN__)
	localtime_s(&vTm, &vSystemTime);
#endif

	// Convert the year, month and day to a Julian day number.
	mvSetDate(vTm.tm_year + 1900, vTm.tm_mon + 1, vTm.tm_mday);
}

///	@brief This function sets a CDate object based on the given year, month and day.
///
///	@param[in]	avYear	The given year.
///	@param[in]	avMonth	The given month.
///	@param[in]	avDay	The given day.
///
///	@exception	std::out_of_range	Indicates the given values are not within the valid ranges.
///
void CDate::mvSetDate(int avYear, int avMonth, int avDay)
{
	// Validate the year, month and day.
	smvValidateDate(avYear, avMonth, avDay);

	// Convert the year, month and day to a Julian day number.
	auto vA = (14 - avMonth) / 12;
	auto vY = avYear + 4800 - vA;
	auto vM = avMonth + 12 * vA - 3;
	auto vDateNo = avDay + ((153 * vM + 2) / 5) + 365 * vY + (vY / 4) - (vY / 100) + (vY / 400) - 32045;

	// Check if the resulting Julian day number is within valid ranges.
	if (vDateNo < MIN_DATENO || vDateNo > MAX_DATENO)
	{
		ostringstream vStream;
		vStream
			<< "Julian day number 'vDateNo' must be within the range " << MIN_DATENO << " to " << MAX_DATENO
			<< " at (" << __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
	mvDateNo = vDateNo;
}

///	@brief This function obtains the equivalent year, month and day of the CDate object.
///
///	@return The equivalent year, month and day of the CDate object.
///
CDate::SDate CDate::mvGetDate() const noexcept
{
	// Convert the Julian day number to year, month and day.
	auto vJ = mvDateNo + 32044;
	auto vG = vJ / 146097;
	auto vDG = vJ % 146097;
	auto vC = ((vDG / 36524 + 1) * 3) / 4;
	auto vDC = vDG - vC * 36524;
	auto vB = vDC / 1461;
	auto vDB = vDC % 1461;
	auto vA = ((vDB / 365 + 1) * 3) / 4;
	auto vDA = vDB - vA * 365;
	auto vY = vG * 400 + vC * 100 + vB * 4 + vA;
	auto vM = (vDA * 5 + 308) / 153 - 2;
	auto vD = vDA - ((vM + 4) * 153) / 5 + 122;

	// Store the results of the conversion.
	SDate vDate;
	vDate.mvYear = vY - 4800 + (vM + 2) / 12;
	vDate.mvMonth = (vM + 2) % 12 + 1;
	vDate.mvDay = vD + 1;

	// Return the result.
	return vDate;
}

///	@brief This function swaps a CDate object with another CDate object.
///
///	@param[in,out]	arRHS	The CDate object to swap with.
///
void CDate::swap(CDate& arRHS) noexcept
{
	using std::swap;
	swap(mvDateNo, arRHS.mvDateNo);
}

///	@brief This function pre-increments a CDate object by one.
///
///	@return	The CDate object.
///
///	@exception	std::out_of_range	Indicates incrementing the CDate object will make the Julian day number go
///									above the maximum supported date.
///
CDate& CDate::operator++()
{
	if ((mvDateNo + 1) > MAX_DATENO)
	{
		ostringstream vStream;
		vStream
			<< "Julian day number 'mvDateNo' exceeds maximum limit " << MAX_DATENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
	++mvDateNo;
	return *this;
}

///	@brief This function post-increments a CDate object by one.
///
///	@return	The CDate object before the increment.
///
CDate CDate::operator++(int)
{
	CDate vDate = *this;
	++(*this);
	return vDate;
}

///	@brief This function pre-decrements a CDate object by one.
///
///	@return	The CDate object.
///
///	@exception	std::out_of_range	Indicates decrementing the CDate object will make the Julian day number go
///									below the minimum supported date.
///
CDate& CDate::operator--()
{
	if ((mvDateNo - 1) < MIN_DATENO)
	{
		ostringstream vStream;
		vStream
			<< "Julian day number 'mvDateNo' exceeds minimum limit " << MIN_DATENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
	--mvDateNo;
	return *this;
}

///	@brief This function post-decrements a CDate object by one.
///
///	@return	The CDate object before the decrement.
///
CDate CDate::operator--(int)
{
	CDate vDate = *this;
	--(*this);
	return vDate;
}

///	@brief This function increments the CDate object by the given duration.
///
///	@param[in]	arDateDuration	The duration to add to the date.
///
///	@return	The CDate object.
///
///	@exception	std::out_of_range	Indicates incrementing the CDate object will make the Julian day number go
///									above the maximum supported date.
///
CDate& CDate::operator+=(const CDateDuration& arDateDuration)
{
	auto vDuration = arDateDuration.mvGetDuration();
	auto vDate = mvGetDate();
	switch (arDateDuration.mvGetUnit())
	{
	case CDateDuration::E_U_YEAR:
		mvSetDate(vDate.mvYear + vDuration, vDate.mvMonth, vDate.mvDay);
		return *this;

	case CDateDuration::E_U_MONTH:
		mvSetDate
		(
			// Adds the overflow years from month calculation.
			vDate.mvYear + (vDate.mvMonth + vDuration - 1) / MONTH_PER_YEAR,
			// Restricts month to within 1 to 12.
			(vDate.mvMonth + vDuration - 1) % MONTH_PER_YEAR + 1,
			vDate.mvDay
		);
		return *this;

	case CDateDuration::E_U_WEEK:
		vDuration *= DAY_PER_WEEK;
		break;

	case CDateDuration::E_U_DAY:
		vDuration *= 1;
		break;

	default:
		break;
	}
	if ((mvDateNo + vDuration) > MAX_DATENO)
	{
		ostringstream vStream;
		vStream
			<< "Julian day number 'mvDateNo' exceeds maximum limit " << MAX_DATENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
	mvDateNo += vDuration;
	return *this;
}

///	@brief This function decrements the CDate object by the given duration.
///
///	@param[in]	arDateDuration	The duration to subtract from the date.
///
///	@return	The CDate object.
///
///	@exception	std::out_of_range	Indicates decrementing the CDate object will make the Julian day number go
///									below the minimum supported date.
///
CDate& CDate::operator-=(const CDateDuration& arDateDuration)
{
	auto vDuration = arDateDuration.mvGetDuration();
	auto vDate = mvGetDate();
	switch (arDateDuration.mvGetUnit())
	{
	case CDateDuration::E_U_YEAR:
		mvSetDate(vDate.mvYear - vDuration, vDate.mvMonth, vDate.mvDay);
		return *this;

	case CDateDuration::E_U_MONTH:
		mvSetDate
		(
			// Subtracts the underflow years from month calculation.
			vDate.mvYear + (vDate.mvMonth - vDuration - MONTH_PER_YEAR) / MONTH_PER_YEAR,
			// Restricts month to within 1 to 12.
			((vDate.mvMonth - vDuration - 1) % MONTH_PER_YEAR + MONTH_PER_YEAR) % MONTH_PER_YEAR + 1,
			vDate.mvDay
		);
		return *this;

	case CDateDuration::E_U_WEEK:
		vDuration *= DAY_PER_WEEK;
		break;

	case CDateDuration::E_U_DAY:
		vDuration *= 1;
		break;

	default:
		break;
	}
	if ((mvDateNo - vDuration) < MIN_DATENO)
	{
		ostringstream vStream;
		vStream
			<< "Julian day number 'mvDateNo' exceeds minimum limit " << MIN_DATENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
	mvDateNo -= vDuration;
	return *this;
}

///	@brief This function receives a CDate object from the given input stream.
///
///	@param[in]	arStream	The source stream.
///
///	@return	The resulting stream state.
///
std::ios_base::iostate CDate::mvInput(std::istream& arStream)
{
	// Obtain the date format.
	auto vFormat = CDateManip::smvGetFormat(arStream);

	// Initialize a local stream state.
	auto vIoState = ios_base::goodbit;

	// Get input from stream.
	istreambuf_iterator<char> vEnd;
	use_facet<CDateGet>(arStream.getloc()).mvGet(arStream, vEnd, arStream, vIoState, *this, vFormat);
	return vIoState;
}

///	@brief This function sends a CDate object to the given output stream.
///
///	@param[in]	arStream	The destination stream.
///
///	@return	The resulting stream state.
///
std::ios_base::iostate CDate::mvOutput(std::ostream& arStream) const
{
	// Obtain the date format.
	auto vFormat = CDateManip::smvGetFormat(arStream);

	// Initialize a local stream state.
	auto vIoState = ios_base::goodbit;

	// Create a string buffer.
	ostringstream vBuffer;

	// Send output to string buffer.
	if (use_facet<CDatePut>(arStream.getloc()).mvPut(vBuffer, *this, vFormat).failed())
	{
		vIoState = ios_base::failbit;
	}

	// Field width adjustment.
	if (vIoState == ios_base::goodbit)
	{
		// Save the contents of the string buffer to a string.
		auto vOutput = vBuffer.str();

		// Determine if we have to do any padding.
		auto vPadding = static_cast<int>(arStream.width() - vOutput.length());
		ostreambuf_iterator<char> vSink {arStream};
		if (vPadding <= 0)
		{
			copy(vOutput.begin(), vOutput.end(), vSink);
		}
		else if (arStream.flags() & ios_base::left)
		{
			copy(vOutput.begin(), vOutput.end(), vSink);
			fill_n(vSink, vPadding, arStream.fill());
		}
		else
		{
			fill_n(vSink, vPadding, arStream.fill());
			copy(vOutput.begin(), vOutput.end(), vSink);
		}
		if (vSink.failed())
		{
			vIoState = ios_base::failbit;
		}
	}
	return vIoState;
}

///	@brief This function allocates an iword/pword index.
///
void CDate::smvAllocateIndex() noexcept
{
	smvIndex = ios_base::xalloc();
}

int CDate::smvIndex;
std::once_flag CDate::smvFlag;

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This function overloads std::swap for CDate objects.
///
///	@param[in,out]	arLHS	The first object to swap.
///	@param[in,out]	arRHS	The second object to swap.
///
void swap(CDate& arLHS, CDate& arRHS) noexcept
{
	arLHS.swap(arRHS);
}

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This function receives a CDate object from an input stream.
///
///	@param[in]	arStream	The source input stream.
///	@param[out]	arDate		The received CDate object.
///
///	@return	The source input stream.
///
std::istream& operator>>(std::istream& arStream, CDate& arDate)
{
	return gvInputGuarded<CDate>(arStream, arDate);
}

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This function sends a CDate object to an output stream.
///
///	@param[in]	arStream	The destination output stream.
///	@param[in]	arDate		The CDate object to send.
///
///	@return	The destination output stream.
///
std::ostream& operator<<(std::ostream& arStream, const CDate& arDate)
{
	return gvOutputGuarded<CDate>(arStream, arDate);
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
