//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include <locale>

#include "lfc1/datetime/cdate.hpp"
#include "lfc1/datetime/clanginfo.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief This function expands the date format by replacing format specifiers with their equivalent format
///		   specifiers.
///
///	@param[in]	avFormat	The date format to be expanded.
///
///	@return	The expanded format string.
///
std::string CDateBase::smvExpandFormat(std::string avFormat)
{
	// Take a copy of the date format.
	string::size_type vPos;
	for (;;)
	{
		if ((vPos = avFormat.find("%D")) != string::npos)
		{
			// Replace %D if it exists in the date format.
			avFormat.replace(vPos, 2, "%m/%d/%y");
		}
		else if ((vPos = avFormat.find("%F")) != string::npos)
		{
			// Replace %F if it exists in the date format.
			avFormat.replace(vPos, 2, "%Y-%m-%d");
		}
		else if ((vPos = avFormat.find("%h")) != string::npos)
		{
			// Replace %h if it exists in the date format.
			avFormat.replace(vPos, 2, "%b");
		}
		else if ((vPos = avFormat.find("%x")) != string::npos)
		{
			// Replace %x if it exists in the date format.
			avFormat.replace(vPos, 2, CLangInfo::smvGetDateFormat());
		}
		else
		{
			break;
		}
	}

	// Return the expanded date format.
	return avFormat;
}

///	@brief This function parses an input iterator for a week day name.
///
///	@param[in]	avNext		The input iterator.
///	@param[in]	avEnd		The input end iterator.
///	@param[in]	arIosBase	The stream formatting information.
///	@param[out]	arIoState	The stream state.
///	@param[out]	arWeekDayNo	The storage for the resulting day of the week.  (0 = Sunday)
///	@param[in]	avIsShort	Short (true) or long (false) week day name indicator.
///
///	@return	The input iterator.
///
CDateBase::TInIt CDateBase::smvGetWeekDayName
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base& arIosBase,
	std::ios_base::iostate& arIoState,
	int& arWeekDayNo,
	bool avIsShort
)
{
	// It's an error if we already have a weekday number.
	if (arWeekDayNo > -1)
	{
		arIoState |= ios_base::failbit;
		return avNext;
	}

	// Parse for weekday name.
	const auto& rCtype = use_facet<ctype<char>>(arIosBase.getloc());
	string vWeekDayName;
	while (avNext != avEnd && rCtype.is(ctype_base::alpha, *avNext))
	{
		vWeekDayName += *avNext;
		++avNext;
	}

	// Find matching weekday name.
	for (auto vI = 0; vI < CDate::DAY_PER_WEEK; ++vI)
	{
		if (avIsShort)
		{
			if (vWeekDayName == CLangInfo::smvGetShortDayName(vI))
			{
				arWeekDayNo = vI;
				break;
			}
		}
		else
		{
			if (vWeekDayName == CLangInfo::smvGetLongDayName(vI))
			{
				arWeekDayNo = vI;
				break;
			}
		}
	}

	// Check if matching abbreviated weekday name found.
	if (arWeekDayNo == -1)
	{
		arIoState |= ios_base::failbit;
	}
	return avNext;
}

///	@brief This function parses an input iterator for a month name.
///
///	@param[in]	avNext		The input iterator.
///	@param[in]	avEnd		The input end iterator.
///	@param[in]	arIosBase	The stream formatting information.
///	@param[out]	arIoState	The stream state.
///	@param[out]	arMonthNo	The storage for the resulting month number.
///	@param[in]	avIsShort	Short (true) or long (false) month name indicator.
///
///	@return	The input iterator.
///
CDateBase::TInIt CDateBase::smvGetMonthName
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base& arIosBase,
	std::ios_base::iostate& arIoState,
	int& arMonthNo,
	bool avIsShort
)
{
	// It's an error if we already have a month number.
	if (arMonthNo > -1)
	{
		arIoState |= ios_base::failbit;
		return avNext;
	}

	// Parse for month name.
	const auto& rCtype = use_facet<ctype<char>>(arIosBase.getloc());
	string vMonthName;
	while (avNext != avEnd && rCtype.is(ctype_base::alpha, *avNext))
	{
		vMonthName += *avNext;
		++avNext;
	}

	// Find matching month name.
	for (auto vI = 1; vI <= CDate::MONTH_PER_YEAR; ++vI)
	{
		if (avIsShort)
		{
			if (vMonthName == CLangInfo::smvGetShortMonthName(vI))
			{
				arMonthNo = vI;
				break;
			}
		}
		else
		{
			if (vMonthName == CLangInfo::smvGetLongMonthName(vI))
			{
				arMonthNo = vI;
				break;
			}
		}
	}

	// Check if matching abbreviated month name found.
	if (arMonthNo == -1)
	{
		arIoState |= ios_base::failbit;
	}
	return avNext;
}

///	@brief This function sends the week day name of a given week day to an output iterator.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	avWeekDayNo	The number of the week day.
///	@param[in]	avIsShort	Short (true) or long (false) week day name indicator.
///
///	@return	The output iterator.
///
CDateBase::TOutIt CDateBase::smvPutWeekDayName(TOutIt avOut, int avWeekDayNo, bool avIsShort)
{
	string vWeekDayName;
	if (avIsShort)
	{
		vWeekDayName = CLangInfo::smvGetShortDayName(avWeekDayNo);
	}
	else
	{
		vWeekDayName = CLangInfo::smvGetLongDayName(avWeekDayNo);
	}
	copy(vWeekDayName.begin(), vWeekDayName.end(), avOut);
	return avOut;
}

///	@brief This function sends the month name of a given month to an output iterator.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	avMonthNo	The number of the month.
///	@param[in]	avIsShort	Short (true) or long (false) month name indicator.
///
///	@return	The output iterator.
///
CDateBase::TOutIt CDateBase::smvPutMonthName(TOutIt avOut, int avMonthNo, bool avIsShort)
{
	string vMonthName;
	if (avIsShort)
	{
		vMonthName = CLangInfo::smvGetShortMonthName(avMonthNo);
	}
	else
	{
		vMonthName = CLangInfo::smvGetLongMonthName(avMonthNo);
	}
	copy(vMonthName.begin(), vMonthName.end(), avOut);
	return avOut;
}

///	@brief This function determines the week-based year and week-based week number of a given date.
///
///	@param[in]		arDate				The given date expressed as a Julian day number.
///	@param[in,out]	arWeekBasedYear		Set to the starting week-based year on input and contains the
///										resulting week-based year on output.
///	@param[out]		apWeekBasedWeekNo	The resulting week-based week number of the given date.  If null,
///										week-based week number is not stored.
///
void CDateBase::smvCalcWeekBasedDetails
(
	const CDate& arDate,
	int& arWeekBasedYear,
	int* apWeekBasedWeekNo
) noexcept
{
	CDate vFirstMonday;
	for (;;)
	{
		// Find the Monday of the week containing the 4th of January.
		vFirstMonday.mvSetDate(arWeekBasedYear, 1, 4);
		while (vFirstMonday % CDate::DAY_PER_WEEK != CDate::FIRST_DAY_MON)
		{
			--vFirstMonday;
		}

		// If the given date is on or after the above Monday then we have found the required week-based year.
		if (arDate >= vFirstMonday)
		{
			break;
		}

		// Change the week-based year.
		--arWeekBasedYear;
	}
	if (apWeekBasedWeekNo != nullptr)
	{
		*apWeekBasedWeekNo = (arDate - vFirstMonday) / CDate::DAY_PER_WEEK + 1;
	}
}

///	@brief This function determines the week number of a given date within a given year.
///
///	@param[in]	arDate	The given date expressed as a Julian day number.
///	@param[in]	avYear	The year of the given date.
///	@param[in]	avDay	The starting day of the week.  (0 or 6 = Monday or Sunday)
///
///	@return The week number of a given date within a given year.
///
int CDateBase::smvCalcWeekNo(const CDate& arDate, int avYear, int avDay) noexcept
{
	CDate vFirstDay {avYear, 1, 1};
	while (vFirstDay % CDate::DAY_PER_WEEK != avDay)
	{
		++vFirstDay;
	}
	return ((arDate - vFirstDay) / CDate::DAY_PER_WEEK + 1);
}

///	@brief This function creates a default CDateBase object.
///
CDateBase::CDateBase() noexcept
{
}

///	@brief This function destroys a CDateBase object.
///
CDateBase::~CDateBase() noexcept
{
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
