//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include "lfc1/datetime/cdate.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief This function obtains the date format from a stream.
///
///	@param[in]	arIosBase	The stream containing the date format.
///
///	@return The date format.
///
std::string CDateManip::smvGetFormat(std::ios_base& arIosBase)
{
	auto pFormat = reinterpret_cast<string*>(arIosBase.pword(CDate::smvGetIndex()));
	if (pFormat == nullptr || pFormat->empty())
	{
		return "%x";
	}
	else
	{
		return *pFormat;
	}
}

///	@brief This function obtains the error information relating to date format copying.
///
///	@param[in]	arIos	The stream containing the error information.
///
///	@return The error information relating to date format copying.
///
std::ios_base::iostate CDateManip::smvCopyFmtErr(std::basic_ios<char>& arIos)
{
	auto vErr = static_cast<ios_base::iostate>(arIos.iword(CDate::smvGetIndex()));
	arIos.iword(CDate::smvGetIndex()) = 0;
	arIos.setstate(vErr | arIos.rdstate());
	return vErr;
}

///	@brief This function creates a CDateManip object.
///
///	@param[in]	apFormat	The desired date format.
///
CDateManip::CDateManip(const char* apFormat)
:	CManipBase<const char*> {smvSetFormat, apFormat}
{
}

///	@brief This function destroys a CDateManip object.
///
CDateManip::~CDateManip() noexcept
{
}

///	@brief This function stores the given date format in a stream.
///
///	@param[in]	arIosBase	The stream containing the date format.
///	@param[in]	apFormat	The date format.
///
///	@exception	std::bad_alloc	Indicates a memory allocation failure.
///
void CDateManip::smvSetFormat(std::ios_base& arIosBase, const char* apFormat)
{
	// Obtain the current value of the pword for date format.
	string*& rPword = reinterpret_cast<string*&>(arIosBase.pword(CDate::smvGetIndex()));

	// Check if the pword is already set or not.
	if (rPword == nullptr)
	{
		// No it is not, register callback function.
		arIosBase.register_callback(smvCallback, CDate::smvGetIndex());
	}
	string* pNewFormat = new string(apFormat);
	string* pOldFormat = rPword;
	rPword = pNewFormat;
	delete pOldFormat;
}

///	@brief This function is the stream callback function.
///
///	@param[in]	avEvent		The stream event.
///	@param[in]	arIosBase	The stream.
///	@param[in]	avIndex		The pword index.
///
void CDateManip::smvCallback(std::ios_base::event avEvent, std::ios_base& arIosBase, int avIndex) noexcept
{
	if (avEvent == ios_base::erase_event)
	{
		try
		{
			delete reinterpret_cast<string*>(arIosBase.pword(avIndex));
		}
		catch (...)
		{
			// Ignore the exception because the stream is about to be destroyed anyway.
		}
	}
	else if (avEvent == ios_base::copyfmt_event)
	{
		string*& rPword = reinterpret_cast<string*&>(arIosBase.pword(avIndex));
		if (rPword != nullptr)
		{
			try
			{
				rPword = new string(*rPword);
			}
			catch (bad_alloc&)
			{
				// Store the error in the stream's iword.  It is the users responsibility to fetch the
				// error stored in the stream's iword after invoking the stream's copyfmt() function.
				arIosBase.iword(avIndex) =
					static_cast<ios_base::iostate>(arIosBase.iword(avIndex)) | ios_base::badbit;
			}
		}
	}
}

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This function is the date manipulator.  It is named following the naming convention of the standard
///		   manipulators.
///
///	@param[in]	apFormat	The date format.  If set to null or an empty string, it clears the date format
///							effectively resetting it to the default date format.
///
///	@return The date manipulator helper.
///
CDateManip setdateformat(const char* apFormat)
{
	return CDateManip {apFormat};
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
