//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include <stdexcept>

#include "lfc1/datetime/cdate.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief The facet ID.
///
std::locale::id CDatePut::id;

///	@brief This function creates a CDatePut object.
///
///	@param[in]	avRelease	Indicates who controls the lifetime of the facet. (0 means locale)
///
CDatePut::CDatePut(size_t avRelease)
:	locale::facet {avRelease}
{
}

///	@brief This function destroys a CDatePut object.
///
CDatePut::~CDatePut() noexcept
{
}

///	@brief This function outputs a date to a stream.
///
///	This function recognizes the following format specifiers which are identical to the date specific format
///	specifiers of the C language strftime() function:
///
///	@code
///	%% - A percent character.
///	%a - Abbreviated weekday name.
///	%A - Full weekday name.
///	%b - Abbreviated month name.
///	%B - Full month name.
///	%C - Century [00-99].
///	%d - Day of the month zero padded [01-31].
///	%D - Equivalent to %m/%d/%y.
///	%e - Day of the month blank padded [ 1-31].
///	%F - Equivalent to %Y-%m-%d.
///	%g - Week-based year without century [00-99].	(See week-based year explanation below.)
///	%G - Week-based year with century [0000-9999].	(See week-based year explanation below.)
///	%h - Equivalent to %b.
///	%j - Day of the year [001-366].
///	%m - Month number [01-12].
///	%n - Newline character.
///	%t - Horizontal tab character.
///	%u - Day of the week [1-7], Monday is 1.
///	%U - Week of the year [00-53], the first Sunday starts week 1.
///	%V - Week number of a week-based year [01-53].	(See week-based year explanation below.)
///	%w - Day of the week [0-6], Sunday is 0.
///	%W - Week of the year [00-53], the first Monday starts week 1.
///	%x - Default date format.
///	%y - Year without century [00-99].
///	%Y - Year with century [0000-9999].
///	@endcode
///
///	Any other sequence of characters not included in the above list will be taken literally.
///
///	The week-based year is the ISO 8601 week-based year.  In this system, weeks begin on a Monday and week 1
///	of the year is the week that includes January 4th, which is also the week that includes the first Thursday
///	of the year and is also the first week that contains at least four days in the year.  If the first Monday
///	of January is the 2nd, 3rd or 4th, the preceding days are part of the last week of the preceding year;
///	thus, for Saturday 2nd January 1999, %G is replaced by 1998 and %V is replaced by 53.  If December 29th,
///	30th or 31st is a Monday, it and any following days are part of week 1 of the following year.  Thus for
///	Tuesday 30th December 1997, %G is replaced by 1998 and %V is replaced by 01.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	arDate		The date to be sent to the output stream.
///	@param[in]	arFormat	The format of the date to be sent to the output stream.
///
///	@return	The output iterator.
///
CDatePut::TOutIt CDatePut::mvPut(TOutIt avOut, const CDate& arDate, const std::string& arFormat) const
{
	return mvDoPut(avOut, arDate, arFormat);
}

///	@brief This function implements the behavior of the mvPut() function.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	arDate		The date to be sent to the output stream.
///	@param[in]	arFormat	The format of the date to be sent to the output stream.
///
///	@return	The output iterator.
///
CDatePut::TOutIt CDatePut::mvDoPut(TOutIt avOut, const CDate& arDate, const std::string& arFormat) const
{
	// Determine the equivalent year, month and day for the given CDate object.
	auto vDate = arDate.mvGetDate();

	// There are some format specifiers which are simply aliases for a combination of other format
	// specifiers.  Replace these format specifiers with their equivalent format specifiers.
	auto vFormat = smvExpandFormat(arFormat);

	// Send output via the iterator.
	string::size_type vFormatOffset {0};
	while (vFormatOffset < vFormat.length())
	{
		// Check for all recognized format specifiers.
		if (vFormat.compare(vFormatOffset, 2, "%%") == 0)
		{
			// Output % character.
			avOut = '%';
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%a") == 0)
		{
			// Output abbreviated weekday name.
			avOut =
				smvPutWeekDayName(avOut, (arDate + CDate::DATENO_OFFSET_SUN) % CDate::DAY_PER_WEEK, true);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%A") == 0)
		{
			// Output full weekday name.
			avOut =
				smvPutWeekDayName(avOut, (arDate + CDate::DATENO_OFFSET_SUN) % CDate::DAY_PER_WEEK, false);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%b") == 0)
		{
			// Output abbreviated month name.
			avOut = smvPutMonthName(avOut, vDate.mvMonth, true);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%B") == 0)
		{
			// Output full month name.
			avOut = smvPutMonthName(avOut, vDate.mvMonth, false);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%C") == 0)
		{
			// Output 2-digit century number zero padded.
			avOut = smvPutNumber2Zero(avOut, vDate.mvYear / 100);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%d") == 0)
		{
			// Output 2-digit day number zero padded.
			avOut = smvPutNumber2Zero(avOut, vDate.mvDay);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%e") == 0)
		{
			// Output 2-digit day number blank padded.
			avOut = smvPutNumber2Blank(avOut, vDate.mvDay);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%g") == 0)
		{
			// Output 2-digit week-based year number without century zero padded.
			int vWeekBasedYear = vDate.mvYear + 1;
			smvCalcWeekBasedDetails(arDate, vWeekBasedYear, nullptr);
			avOut = smvPutNumber2Zero(avOut, vWeekBasedYear % 100);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%G") == 0)
		{
			// Output 4-digit week-based year number with century zero padded.
			int vWeekBasedYear = vDate.mvYear + 1;
			smvCalcWeekBasedDetails(arDate, vWeekBasedYear, nullptr);
			avOut = smvPutNumber4Zero(avOut, vWeekBasedYear);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%j") == 0)
		{
			// Output 3-digit day of the year number zero padded.
			CDate vStartYear {vDate.mvYear, 1, 1};
			avOut = smvPutNumber3Zero(avOut, arDate - vStartYear + 1);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%m") == 0)
		{
			// Output 2-digit month number zero padded.
			avOut = smvPutNumber2Zero(avOut, vDate.mvMonth);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%n") == 0)
		{
			// Output newline character.
			avOut = '\n';
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%t") == 0)
		{
			// Output horizontal tab character.
			avOut = '\t';
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%u") == 0)
		{
			// Output 1-digit weekday number, Monday is 1.
			avOut = smvPutNumber1(avOut, arDate % CDate::DAY_PER_WEEK + 1);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%U") == 0)
		{
			// Output 2-digit week number zero padded with weeks beginning on Sunday.
			int vWeekNo = smvCalcWeekNo(arDate, vDate.mvYear, CDate::FIRST_DAY_SUN);
			avOut = smvPutNumber2Zero(avOut, vWeekNo);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%V") == 0)
		{
			// Output 2-digit week number of a week-based year zero padded.
			int vWeekBasedYear = vDate.mvYear + 1;
			int vWeekBasedWeekNo;
			smvCalcWeekBasedDetails(arDate, vWeekBasedYear, &vWeekBasedWeekNo);
			avOut = smvPutNumber2Zero(avOut, vWeekBasedWeekNo);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%w") == 0)
		{
			// Output 1-digit weekday number, Sunday is 0.
			avOut = smvPutNumber1(avOut, (arDate + CDate::DATENO_OFFSET_SUN) % CDate::DAY_PER_WEEK);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%W") == 0)
		{
			// Output 2-digit week number zero padded with weeks beginning on Monday.
			int vWeekNo = smvCalcWeekNo(arDate, vDate.mvYear, CDate::FIRST_DAY_MON);
			avOut = smvPutNumber2Zero(avOut, vWeekNo);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%y") == 0)
		{
			// Output 2-digit year number without century zero padded.
			avOut = smvPutNumber2Zero(avOut, vDate.mvYear % 100);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%Y") == 0)
		{
			// Output 4-digit year number with century zero padded.
			avOut = smvPutNumber4Zero(avOut, vDate.mvYear);
			vFormatOffset += 2;
		}
		else
		{
			// Output normal character.
			avOut = vFormat[vFormatOffset];
			++vFormatOffset;
		}
	}
	return avOut;
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
