//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include <algorithm>
#include <sstream>
#include <stdexcept>
#include <ctime>
#include <cstring>
#include <system_error>
#if defined(__LINUX__) || defined(__UNIX__)
#include <sys/types.h>
#include <sys/timeb.h>
#endif
#if defined(__WIN__)
#include <sys\types.h>
#include <sys\timeb.h>
#endif

#include "lfc1/datetime/cdatetime.hpp"
#include "lfc1/datetime/cdate.hpp"
#include "lfc1/datetime/ctime.hpp"
#include "lfc1/udtsup/input.hpp"
#include "lfc1/udtsup/output.hpp"

using namespace std;
using namespace lfc1::udtsup;

namespace lfc1 { namespace datetime {

///	@brief This function obtains an iword/pword index.
///
///	@return	The iword/pword index.
///
int CDateTime::smvGetIndex() noexcept
{
	call_once(smvFlag, smvAllocateIndex);
	return smvIndex;
}

///	@brief This function validates a daylight savings indicator.
///
///	@param[in]	avDst	The daylight savings indicator (-1, 0 or 1).
///
///	@exception	std::out_of_range	Indicates that the daylight savings indicator has an invalid value.
///
void CDateTime::smvValidateDst(int avDst)
{
	if (avDst < -1 || avDst > 1)
	{
		ostringstream vStream;
		vStream
			<< "Daylight savings indicator 'avDst' must be in the range -1 to 1 at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
}

///	@brief This function creates a default CDateTime object.
///
CDateTime::CDateTime() noexcept
:	mvDateTimeNo {0}
{
	mvSetDateTime();
}

///	@brief This function creates a CDateTime object using the given number of milliseconds since
///		   01-Jan-1970 00:00:00.000.
///
///	@param[in]	avDateTimeNo	The number of milliseconds since 01-Jan-1970 00:00:00.000.
///
///	@exception	std::out_of_range	Indicates that the local date and time represented by the number of
///									milliseconds since 01-Jan-1970 00:00:00.000 UTC is not within valid range.
///
CDateTime::CDateTime(long long avDateTimeNo)
:	mvDateTimeNo {avDateTimeNo}
{
	if (avDateTimeNo < MIN_DATETIMENO || avDateTimeNo > MAX_DATETIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Julian day number 'avDateTimeNo' must be within the range " << MIN_DATETIMENO << " to "
			<< MAX_DATETIMENO << " at (" << __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
}

///	@brief This function creates a CDateTime object using the given year, month, day, hour, minute, second,
///		   millisecond and daylight savings indicator.
///
///	@param[in]	avYear			The given year.
///	@param[in]	avMonth			The given month.
///	@param[in]	avDay			The given day.
///	@param[in]	avHour			The given hour.
///	@param[in]	avMinute		The given minute.
///	@param[in]	avSecond		The given second.
///	@param[in]	avMillisecond	The given millisecond.
///	@param[in]	avDst			The given daylight savings indicator.  (-1 means daylight state unknown, 0
///								means date and time is a standard date and time and 1 means date and time is
///								a daylight date and time.
///
CDateTime::CDateTime
(
	int avYear,
	int avMonth,
	int avDay,
	int avHour,
	int avMinute,
	int avSecond,
	int avMillisecond,
	int avDst
)
:	mvDateTimeNo {0}
{
	mvSetDateTime(avYear, avMonth, avDay, avHour, avMinute, avSecond, avMillisecond, avDst);
}

///	@brief This function copy constructs a CDateTime object.
///
///	@param[in]	arRHS	The CDateTime object to be copied.
///
CDateTime::CDateTime(const CDateTime& arRHS) noexcept
:	mvDateTimeNo {arRHS.mvDateTimeNo}
{
}

///	@brief This function move constructs a CDateTime object.
///
///	@param[in]	arRHS	The CDateTime object to be moved.
///
CDateTime::CDateTime(CDateTime&& arRHS) noexcept
:	mvDateTimeNo {arRHS.mvDateTimeNo}
{
}

///	@brief This function destroys a CDateTime object.
///
CDateTime::~CDateTime() noexcept
{
}

///	@brief This function assigns the given number of milliseconds since 01-Jan-1970 00:00:00.000 to a
///		   CDateTime object.
///
///	@param[in]	avDateTimeNo	The given number of milliseconds since 01-Jan-1970 00:00:00.000.
///
///	@return	The CDateTime object assigned to.
///
CDateTime& CDateTime::operator=(long long avDateTimeNo)
{
	CDateTime vCopy {avDateTimeNo};
	swap(vCopy);
	return *this;
}

///	@brief This function assigns a CDateTime object to another CDateTime object.
///
///	@param[in]	arRHS	The CDateTime object to assign to another CDateTime object.
///
///	@return	The CDateTime object assigned to.
///
CDateTime& CDateTime::operator=(const CDateTime& arRHS) noexcept
{
	mvDateTimeNo = arRHS.mvDateTimeNo;
	return *this;
}

///	@brief This function moves a CDateTime object to another CDateTime object.
///
///	@param[in]	arRHS	The CDateTime object to move to another CDateTime object.
///
///	@return	The CDateTime object assigned to.
///
CDateTime& CDateTime::operator=(CDateTime&& arRHS) noexcept
{
	mvDateTimeNo = arRHS.mvDateTimeNo;
	return *this;
}

///	@brief This function obtains the number of milliseconds since 01-Jan-1970 00:00:00.000.
///
///	@return	The number of milliseconds since 01-Jan-1970 00:00:00.000.
///
CDateTime::operator long long () const noexcept
{
	return mvDateTimeNo;
}

///	@brief This function assigns the current system time to a CDateTime object.
///
void CDateTime::mvSetDateTime() noexcept
{
	// Obtain the current system time with millisecond precision.
#if defined(__LINUX__) || defined(__UNIX__)
	timeb vNow;
	ftime(&vNow);
#endif
#if defined(__WIN__)
	_timeb vNow;
	_ftime_s(&vNow);
#endif
	auto vSystemTime = vNow.time;
	auto vMillisecond = static_cast<int>(vNow.millitm);

	// Convert the current system time to a structure.
	tm vTm;
	memset(&vTm, 0, sizeof(vTm));
#if defined(__LINUX__) || defined(__UNIX__)
	localtime_r(&vSystemTime, &vTm);
#endif
#if defined(__WIN__)
	localtime_s(&vTm, &vSystemTime);
#endif
	mvSetDateTime
	(
		vTm.tm_year + 1900,
		vTm.tm_mon + 1,
		vTm.tm_mday,
		vTm.tm_hour,
		vTm.tm_min,
		vTm.tm_sec,
		vMillisecond
	);
}

///	@brief This function sets a CDateTime object based on the given year, month, day, hour, minute, second,
///		   millisecond and daylight savings indicator.
///
///	@param[in]	avYear			The given year.
///	@param[in]	avMonth			The given month.
///	@param[in]	avDay			The given day.
///	@param[in]	avHour			The given hour.
///	@param[in]	avMinute		The given minute.
///	@param[in]	avSecond		The given second.
///	@param[in]	avMillisecond	The given millisecond.
///	@param[in]	avDst			The given daylight savings indicator.  (-1 means daylight state unknown, 0
///								means date and time is a standard date and time and 1 means date and time is
///								a daylight date and time.
///
///	@exception	std::system_error	Indicates that conversion to a time_t value failed.
///	@exception	std::out_of_range	Indicates that the number of milliseconds since 01-Jan-1970 00:00:00.000
///									is out-of-range.
///
void CDateTime::mvSetDateTime
(
	int avYear,
	int avMonth,
	int avDay,
	int avHour,
	int avMinute,
	int avSecond,
	int avMillisecond,
	int avDst
)
{
	// Validate the year, month, day, hour, minute, second, millisecond and daylight savings indicator.
	CDate::smvValidateDate(avYear, avMonth, avDay);
	CTime::smvValidateTime(avHour, avMinute, avSecond, avMillisecond);
	smvValidateDst(avDst);

	// Convert the given values to a system time.
	tm vTm;
	memset(&vTm, 0, sizeof(vTm));
	vTm.tm_year = avYear - 1900;
	vTm.tm_mon = avMonth - 1;
	vTm.tm_mday = avDay;
	vTm.tm_hour = avHour;
	vTm.tm_min = avMinute;
	vTm.tm_sec = avSecond;
	vTm.tm_isdst = avDst;
	errno = 0;
	auto vSystemTime = mktime(&vTm);
	if (vSystemTime < 0 && errno != 0)
	{
		ostringstream vStream;
		vStream << "mktime() failed at (" << __FILE__ << ", " << __LINE__ << ")";
		throw system_error {errno, generic_category(), vStream.str()};
	}
	auto vDateTimeNo = static_cast<long long>(vSystemTime) * CTime::MSEC_PER_SEC + avMillisecond;

	// Check if the resulting milliseconds since 01-Jan-1970 00:00:00.000 is within valid ranges.
	if (vDateTimeNo < MIN_DATETIMENO || vDateTimeNo > MAX_DATETIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Number of milliseconds 'vDateTimeNo' must be within the range " << MIN_DATETIMENO << " to "
			<< MAX_DATETIMENO << " at (" << __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
	mvDateTimeNo = vDateTimeNo;
}

///	@brief This function obtains the equivalent local year, month, day, hour, minute, second, millisecond and
///		   daylight saving time indicator from a CDateTimeNo object.
///
///	@return The equivalent local year, month, day, hour, minute, second, millisecond and  daylight saving time
///			indicator from a CDateTimeNo object.
///
CDateTime::SDateTime CDateTime::mvGetDateTime() const noexcept
{
	// Convert the number of milliseconds since 01-Jan-1970 00:00:00.000 UTC  to system time and milliseconds.
	auto vSystemTime = static_cast<time_t>(mvDateTimeNo / CTime::MSEC_PER_SEC);
	auto vMillisecond = static_cast<int>(mvDateTimeNo % CTime::MSEC_PER_SEC);

	// Convert the system time to a structure.
	tm vTm;
	memset(&vTm, 0, sizeof(vTm));
#if defined(__LINUX__) || defined(__UNIX__)
	localtime_r(&vSystemTime, &vTm);
#endif
#if defined(__WIN__)
	localtime_s(&vTm, &vSystemTime);
#endif

	// Store the results of the conversions.
	SDateTime vDateTime;
	vDateTime.mvYear = vTm.tm_year + 1900;
	vDateTime.mvMonth = vTm.tm_mon + 1;
	vDateTime.mvDay = vTm.tm_mday;
	vDateTime.mvHour = vTm.tm_hour;
	vDateTime.mvMinute = vTm.tm_min;
	vDateTime.mvSecond = vTm.tm_sec;
	vDateTime.mvMillisecond = vMillisecond;
	vDateTime.mvDst = vTm.tm_isdst;

	// Return the result.
	return vDateTime;
}

///	@brief This function swaps a CDateTime object with another CDateTime object.
///
///	@param[in,out]	arRHS	The CDateTime object to swap with.
///
void CDateTime::swap(CDateTime& arRHS) noexcept
{
	using std::swap;
	swap(mvDateTimeNo, arRHS.mvDateTimeNo);
}

///	@brief This function pre-increments a CDateTime object by one.
///
///	@return	The CDateTime object.
///
///	@exception	std::out_of_range	Indicates incrementing the number of milliseconds since 01-Jan-1970
///									00:00:00.000 UTC will make the date go above the maximum supported date.
///
CDateTime& CDateTime::operator++()
{
	if ((mvDateTimeNo + 1) > MAX_DATETIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Number of milliseconds 'mvDateTimeNo' exceeds maximum limit " << MAX_DATETIMENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
	++mvDateTimeNo;
	return *this;
}

///	@brief This function post-increments a CDateTime object by one.
///
///	@return	The CDateTime object before the increment.
///
CDateTime CDateTime::operator++(int)
{
	CDateTime vDateTimeNo = *this;
	++(*this);
	return vDateTimeNo;
}

///	@brief This function pre-decrements a CDateTime object by one.
///
///	@return	The CDateTime object.
///
///	@exception	std::out_of_range	Indicates decrementing the number of milliseconds since 01-Jan-1970
///									00:00:00.000 UTC will make the date go below the minimum supported date.
///
CDateTime& CDateTime::operator--()
{
	if ((mvDateTimeNo - 1) < MIN_DATETIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Number of milliseconds 'mvDateTimeNo' exceeds minimum limit " << MIN_DATETIMENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
	--mvDateTimeNo;
	return *this;
}

///	@brief This function post-decrements a CDateTime object by one.
///
///	@return	The CDateTime object before the decrement.
///
CDateTime CDateTime::operator--(int)
{
	CDateTime vDateTimeNo = *this;
	--(*this);
	return vDateTimeNo;
}

///	@brief This function iIncrements the CDateTime object by the given duration.
///
///	@param[in]	arDateTimeDuration	The duration to add to the date and time.
///
///	@return	The CDateTime object.
///
///	@exception	std::out_of_range	Indicates incrementing the date and time will make the date and time go
///									above the maximum supported date and time.
///
CDateTime& CDateTime::operator+=(const CDateTimeDuration& arDateTimeDuration)
{
	auto vDuration1 = arDateTimeDuration.mvGetDuration();
	long long vDuration2 = arDateTimeDuration.mvGetDuration();

	// Add the duration to the CDateTime object.
	auto vDateTime = mvGetDateTime();
	switch (arDateTimeDuration.mvGetUnit())
	{
	case CDateTimeDuration::E_U_YEAR:
		mvSetDateTime
		(
			vDateTime.mvYear + vDuration1,
			vDateTime.mvMonth,
			vDateTime.mvDay,
			vDateTime.mvHour,
			vDateTime.mvMinute,
			vDateTime.mvSecond,
			vDateTime.mvMillisecond
		);
		return *this;

	case CDateTimeDuration::E_U_MONTH:
		mvSetDateTime
		(
			// Adds the overflow years from month calculation.
			vDateTime.mvYear + (vDateTime.mvMonth + vDuration1 - 1) / CDate::MONTH_PER_YEAR,
			// Restricts month to within 1 to 12.
			(vDateTime.mvMonth + vDuration1 - 1) % CDate::MONTH_PER_YEAR + 1,
			vDateTime.mvDay,
			vDateTime.mvHour,
			vDateTime.mvMinute,
			vDateTime.mvSecond,
			vDateTime.mvMillisecond
		);
		return *this;

	case CDateTimeDuration::E_U_WEEK:
		vDuration2 *= CDate::DAY_PER_WEEK * CTime::MSEC_PER_DAY;
		break;

	case CDateTimeDuration::E_U_DAY:
		vDuration2 *= CTime::MSEC_PER_DAY;
		break;

	case CDateTimeDuration::E_U_HOUR:
		vDuration2 *= CTime::MSEC_PER_HR;
		break;

	case CDateTimeDuration::E_U_MINUTE:
		vDuration2 *= CTime::MSEC_PER_MIN;
		break;

	case CDateTimeDuration::E_U_SECOND:
		vDuration2 *= CTime::MSEC_PER_SEC;
		break;

	case CDateTimeDuration::E_U_MILLISECOND:
		vDuration2 *= 1;
		break;

	default:
		break;
	}

	// Check for overflow.
	if ((mvDateTimeNo + vDuration2) > MAX_DATETIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Number of milliseconds 'mvDateTimeNo' exceeds maximum limit " << MAX_DATETIMENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}

	// Add the duration to the CTime object.
	mvDateTimeNo += vDuration2;

	// Return the incremented value.
	return *this;
}

///	@brief This function decrements the CDateTime object by the given duration.
///
///	@param[in]	arDateTimeDuration	The duration to subtract from the date and time.
///
///	@return	The CDateTime object.
///
///	@exception	std::out_of_range	Indicates decrementing the date and time will make the date and time go
///									below the minimum supported date and time.
///
CDateTime& CDateTime::operator-=(const CDateTimeDuration& arDateTimeDuration)
{
	auto vDuration1 = arDateTimeDuration.mvGetDuration();
	long long vDuration2 = arDateTimeDuration.mvGetDuration();

	// Subtract the duration from the CDateTime object.
	auto vDateTime = mvGetDateTime();
	switch (arDateTimeDuration.mvGetUnit())
	{
	case CDateTimeDuration::E_U_YEAR:
		mvSetDateTime
		(
			vDateTime.mvYear - vDuration1,
			vDateTime.mvMonth,
			vDateTime.mvDay,
			vDateTime.mvHour,
			vDateTime.mvMinute,
			vDateTime.mvSecond,
			vDateTime.mvMillisecond
		);
		return *this;

	case CDateTimeDuration::E_U_MONTH:
		mvSetDateTime
		(
			// Subtracts the underflow years from month calculation.
			vDateTime.mvYear +
				(vDateTime.mvMonth - vDuration1 - CDate::MONTH_PER_YEAR) / CDate::MONTH_PER_YEAR,
			// Restricts month to within 1 to 12.
			(
				(vDateTime.mvMonth - vDuration1 - 1) % CDate::MONTH_PER_YEAR + CDate::MONTH_PER_YEAR
			) % CDate::MONTH_PER_YEAR + 1,
			vDateTime.mvDay,
			vDateTime.mvHour,
			vDateTime.mvMinute,
			vDateTime.mvSecond,
			vDateTime.mvMillisecond
		);
		return *this;

	case CDateTimeDuration::E_U_WEEK:
		vDuration2 *= CDate::DAY_PER_WEEK * CTime::MSEC_PER_DAY;
		break;

	case CDateTimeDuration::E_U_DAY:
		vDuration2 *= CTime::MSEC_PER_DAY;
		break;

	case CDateTimeDuration::E_U_HOUR:
		vDuration2 *= CTime::MSEC_PER_HR;
		break;

	case CDateTimeDuration::E_U_MINUTE:
		vDuration2 *= CTime::MSEC_PER_MIN;
		break;

	case CDateTimeDuration::E_U_SECOND:
		vDuration2 *= CTime::MSEC_PER_SEC;
		break;

	case CDateTimeDuration::E_U_MILLISECOND:
		vDuration2 *= 1;
		break;

	default:
		break;
	}

	// Check for underflow.
	if ((mvDateTimeNo - vDuration2) < MIN_DATETIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Number of milliseconds 'mvDateTimeNo' exceeds minimum limit " << MIN_DATETIMENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}

	// Subtract the duration from the CTime object.
	mvDateTimeNo -= vDuration2;

	// Return the decremented value.
	return *this;
}

///	@brief This function receives a CDateTime object from the given input stream.
///
///	@param[in]	arStream	The source stream.
///
///	@return	The resulting stream state.
///
std::ios_base::iostate CDateTime::mvInput(std::istream& arStream)
{
	// Obtain the date and time format.
	auto vFormat = CDateTimeManip::smvGetFormat(arStream);

	// Initialize a local stream state.
	auto vIoState = ios_base::goodbit;

	// Get input from stream.
	istreambuf_iterator<char> vEnd;
	use_facet<CDateTimeGet>(arStream.getloc()).mvGet(arStream, vEnd, arStream, vIoState, *this, vFormat);
	return vIoState;
}

///	@brief This function sends a CDateTime object to the given output stream.
///
///	@param[in]	arStream	The destination stream.
///
///	@return	The resulting stream state.
///
std::ios_base::iostate CDateTime::mvOutput(std::ostream& arStream) const
{
	// Obtain the date and time format.
	auto vFormat = CDateTimeManip::smvGetFormat(arStream);

	// Initialize a local stream state.
	auto vIoState = ios_base::goodbit;

	// Create a string buffer.
	ostringstream vBuffer;

	// Send output to string buffer.
	if (use_facet<CDateTimePut>(arStream.getloc()).mvPut(vBuffer, *this, vFormat).failed())
	{
		vIoState = ios_base::badbit;
	}

	// Field width adjustment.
	if (vIoState == ios_base::goodbit)
	{
		// Save the contents of the string buffer to a string.
		auto vOutput = vBuffer.str();

		// Determine if we have to do any padding.
		auto vPadding = static_cast<int>(arStream.width() - vOutput.length());
		ostreambuf_iterator<char> vSink {arStream};
		if (vPadding <= 0)
		{
			copy(vOutput.begin(), vOutput.end(), vSink);
		}
		else if (arStream.flags() & ios_base::left)
		{
			copy(vOutput.begin(), vOutput.end(), vSink);
			fill_n(vSink, vPadding, arStream.fill());
		}
		else
		{
			fill_n(vSink, vPadding, arStream.fill());
			copy(vOutput.begin(), vOutput.end(), vSink);
		}
		if (vSink.failed())
		{
			vIoState = ios_base::failbit;
		}
	}
	return vIoState;
}

///	@brief This function allocates an iword/pword index.
///
void CDateTime::smvAllocateIndex() noexcept
{
	smvIndex = ios_base::xalloc();
}

int CDateTime::smvIndex;
std::once_flag CDateTime::smvFlag;

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdatetime.hpp
///	@brief This function overloads std::swap for CDateTime objects.
///
///	@param[in,out]	arLHS	The first object to swap.
///	@param[in,out]	arRHS	The second object to swap.
///
void swap(CDateTime& arLHS, CDateTime& arRHS) noexcept
{
	arLHS.swap(arRHS);
}

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdatetime.hpp
///	@brief This function receives a CDateTime object from an input stream.
///
///	@param[in]	arStream	The source input stream.
///	@param[out]	arDateTime		The received CDateTime object.
///
///	@return	The source input stream.
///
std::istream& operator>>(std::istream& arStream, CDateTime& arDateTime)
{
	return gvInputGuarded<CDateTime>(arStream, arDateTime);
}

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdatetime.hpp
///	@brief This function sends a CDateTime object to an output stream.
///
///	@param[in]	arStream	The destination output stream.
///	@param[in]	arDateTime		The CDateTime object to send.
///
///	@return	The destination output stream.
///
std::ostream& operator<<(std::ostream& arStream, const CDateTime& arDateTime)
{
	return gvOutputGuarded<CDateTime>(arStream, arDateTime);
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
