//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include "lfc1/datetime/cdatetime.hpp"
#include "lfc1/datetime/clanginfo.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief This function expands the date and time format by replacing format specifiers with their equivalent
///		   format specifiers.
///
///	@param[in]	avFormat	The date and time format to be expanded.
///
///	@return	The expanded format string.
///
std::string CDateTimeBase::smvExpandFormat(std::string avFormat)
{
	// Perform expansions by the base classes.
	string::size_type vPos;
	for (;;)
	{
		avFormat = CDateBase::smvExpandFormat(avFormat);
		avFormat = CTimeBase::smvExpandFormat(avFormat);
		if ((vPos = avFormat.find("%c")) != string::npos)
		{
			// Replace %c if it exists in the date and time format.
			avFormat.replace(vPos, 2, CLangInfo::smvGetDateTimeFormat());
		}
		else
		{
			break;
		}
	}

	// Return the expanded date and time format.
	return avFormat;
}

///	@brief This function parses an input iterator for a timezone name.
///
///	@param[in]	avNext		The input iterator.
///	@param[in]	avEnd		The input end iterator.
///	@param[out]	arIoState	The stream state.
///	@param[out]	arDst		The storage for the resulting daylight saving time indicator.
///
///	@return	The input iterator.
///
CDateTimeBase::TInIt CDateTimeBase::smvGetTzName
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base::iostate& arIoState,
	int& arDst
)
{
	// It's an error if a daylight saving time indicator already exists.
	if (arDst != -1)
	{
		arIoState |= ios_base::failbit;
		return avNext;
	}

	// Get standard and daylight timezone names.
	string vStandardName;
	string vDaylightName;
#if defined(__LINUX__) || defined(__UNIX__)
	tzset();
	vStandardName = tzname[0];
	vDaylightName = tzname[1];
#endif
#if defined(__WIN__)
	_tzset();
	vStandardName = _tzname[0];
	vDaylightName = _tzname[1];
#endif

	// Parse for timezone name.
	string vName;
	string::size_type vOffset {0};
	auto vDst = -1;
	while (avNext != avEnd)
	{
		if
		(
			vDst == -1
			&&
			vOffset < vStandardName.length() && *avNext == vStandardName[vOffset]
			&&
			vOffset < vDaylightName.length() && *avNext == vDaylightName[vOffset]
		)
		{
			// This is when the first few characters of standard and daylight name are identical.
			++vOffset;
		}
		else if
		(
			(vDst == -1 || vDst == 0)
			&&
			vOffset < vStandardName.length() && *avNext == vStandardName[vOffset]
		)
		{
			// This is when we have matched the first few characters of standard name.
			++vOffset;
			vDst = 0;
		}
		else if
		(
			(vDst == -1 || vDst == 1)
			&&
			vOffset < vDaylightName.length() && *avNext == vDaylightName[vOffset]
		)
		{
			// This is when we have matched the first few characters of daylight name.
			++vOffset;
			vDst = 1;
		}
		else
		{
			break;
		}
		vName += *avNext;
		++avNext;
	}

	// Check if we have matched either the standard or the daylight name.
	if
	(
		(vDst == 0 && vOffset == vStandardName.length())
		||
		(vDst == 1 && vOffset == vDaylightName.length())
	)
	{
		// Standard or daylight name has been matched.
		arDst = vDst;
	}
	else if (vDst == -1 && vName == vStandardName)
	{
		// The standard and the daylight name are identical and we have matched it.
		arDst = 0;
	}
	else
	{
		// We have not matched either the standard or the daylight name.
		arIoState |= ios_base::failbit;
	}
	return avNext;
}

///	@brief This function parses an input iterator for a timezone name.
///
///	@param[in]	avNext		The input iterator.
///	@param[in]	avEnd		The input end iterator.
///	@param[in]	arIosBase	The stream formatting information.
///	@param[out]	arIoState	The stream state.
///	@param[out]	arDst		The storage for the resulting daylight saving time indicator.
///
///	@return	The input iterator.
///
CDateTimeBase::TInIt CDateTimeBase::smvGetTzOffset
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base& arIosBase,
	std::ios_base::iostate& arIoState,
	int& arDst
)
{
	// It's an error if a daylight saving time indicator already exists.
	if (arDst != -1)
	{
		arIoState |= ios_base::failbit;
		return avNext;
	}

	// Input '+' or '-' sign.
	if (avNext == avEnd)
	{
		arIoState |= ios_base::failbit;
		return avNext;
	}
	auto vSign = 0;
	if (*avNext == '+')
	{
		vSign = 1;
	}
	else if (*avNext == '-')
	{
		vSign = -1;
	}
	else
	{
		arIoState |= ios_base::failbit;
		return avNext;
	}
	++avNext;

	// Input the number of hours.
	auto vHour = -1;
	avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vHour, 2, '0');
	if (arIoState != ios_base::goodbit)
	{
		return avNext;
	}

	// Input the number of minutes.
	auto vMinute = -1;
	avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vMinute, 2, '0');
	if (arIoState != ios_base::goodbit)
	{
		return avNext;
	}

	// Combine the sign, hour and minute into a timezone offset.
	auto vOffset = vSign * (vHour * CTime::SEC_PER_HR + vMinute * CTime::SEC_PER_MIN);

	// Compare the entered timezone offset with the actual timezone offset.
#if defined(__LINUX__) || defined(__UNIX__)
	tzset();
	if (vOffset == -timezone)
#endif
#if defined(__WIN__)
	_tzset();
	if (vOffset == -_timezone)
#endif
	{
		arDst = 0;
	}
	else
	{
		arDst = 1;
	}

	return avNext;
}

///	@brief This function sends the timezone name to an output iterator.
///
///	@param[in]	avOut	The output iterator.
///	@param[in]	avDst	The daylight saving time indicator.
///
///	@return The output iterator.
///
CDateTimeBase::TOutIt CDateTimeBase::smvPutTzName(TOutIt avOut, int avDst)
{
	// Get the standard and the daylight timezone names.
	string vStandardName;
	string vDaylightName;
#if defined(__LINUX__) || defined(__UNIX__)
	tzset();
	vStandardName = tzname[0];
	vDaylightName = tzname[1];
#endif
#if defined(__WIN__)
	_tzset();
	vStandardName = _tzname[0];
	vDaylightName = _tzname[1];
#endif

	// Send the correct name to the output iterator.
	string vName;
	if (avDst == 0)
	{
		vName = vStandardName;
	}
	else if (avDst == 1)
	{
		vName = vDaylightName;
	}
	copy(vName.begin(), vName.end(), avOut);
	return avOut;
}

///	@brief This function sends the timezone offset to an output iterator.
///
///	@param[in]	avOut	The output iterator.
///	@param[in]	avDst	The daylight saving time indicator.
///
///	@return The output iterator.
///
CDateTimeBase::TOutIt CDateTimeBase::smvPutTzOffset(TOutIt avOut, int avDst) noexcept
{
#if defined(__LINUX__) || defined(__UNIX__)
	tzset();
	auto vOffset = -timezone;
#endif
#if defined(__WIN__)
	_tzset();
	auto vOffset = -_timezone;
#endif
	if (avDst == 1)
	{
		vOffset += CTime::SEC_PER_HR;
	}
	if (avDst == 0 || avDst == 1)
	{
		// Output '+' or '-' sign.
		avOut = ((vOffset >= 0) ? '+' : '-');

		// Output the number of hours.
		avOut = smvPutNumber2Zero(avOut, abs(vOffset) / CTime::SEC_PER_HR);

		// Output the number of minutes.
		avOut = smvPutNumber2Zero(avOut, (abs(vOffset) % CTime::SEC_PER_HR) / CTime::SEC_PER_MIN);
	}
	return avOut;
}

///	@brief Creates a default CDateTimeBase object.
///
CDateTimeBase::CDateTimeBase() noexcept
{
}

///	@brief Destroys a CDateTimeBase object.
///
CDateTimeBase::~CDateTimeBase() noexcept
{
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
