//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include "lfc1/datetime/cdatetime.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief The facet ID.
///
std::locale::id CDateTimeGet::id;

///	@brief This function creates a CDateTimeGet object.
///
///	@param[in]	avRelease	Indicates who controls the lifetime of the facet. (0 means locale)
///
CDateTimeGet::CDateTimeGet(size_t avRelease)
:	locale::facet {avRelease}
{
}

///	@brief This function destroys a CDateTimeGet object.
///
CDateTimeGet::~CDateTimeGet() noexcept
{
}

///	@brief This function inputs a date and time from a stream.  Duplication of date or time parts is not
///		   allowed.
///
///	This function recognizes the following format specifiers which are identical to the time specific format
///	specifiers of the C language strftime() function except for %N:
///
///	@code
///	%% - A percent character.
///	%a - Abbreviated weekday name.
///	%A - Full weekday name.
///	%b - Abbreviated month name.
///	%B - Full month name.
///	%c - Default date and time format.
///	%C - Century [00-99].
///	%d - Day of the month zero padded [01-31].
///	%D - Equivalent to %m/%d/%y.
///	%e - Day of the month blank padded [ 1-31].
///	%F - Equivalent to %Y-%m-%d.
///	%g - Week-based year without century [00-99].
///	%G - Week-based year with century [0000-9999].
///	%h - Equivalent to %b.
///	%H - Hour in 24-hour format [00-23] zero padded.
///	%I - Hour in 12-hour format [01-12] zero padded.
///	%j - Day of the year [001-366].
///	%k - Hour in 24-hour format [00-23] blank padded.
///	%l - Hour in 12-hour format [01-12] blank padded.
///	%m - Month number [01-12].
///	%M - Minute [00-59].
///	%n - Newline character.
///	%N - Millisecond [000-999].
///	%p - AM/PM designation.
///	%r - Default 12-hour time format.
///	%R - Equivalent to %H:%M.
///	%S - Second [00-59].
///	%t - Horizontal tab character.
///	%T - Equivalent to %H:%M:%S.
///	%u - Day of the week [1-7], Monday is 1.
///	%U - Week of the year [00-53], the first Sunday starts week 1.
///	%V - Week number of a week-based year [01-53].
///	%w - Day of the week [0-6], Sunday is 0.
///	%W - Week of the year [00-53], the first Monday starts week 1.
///	%x - Default date format.
///	%X - Default time format.
///	%y - Year without century [00-99].
///	%Y - Year with century [0000-9999].
///	%z - Time zone offset.	(Accepted but not used.)
///	%Z - Time zone name.	(Accepted but not used.)
///	@endcode
///
///	Any other sequence of characters not included in the above list will be taken literally.
///
///	The following combinations of format specifiers results into a valid date for a CDateTime object:
///
///	@code
///	%D
///	%F
///	%x
///	(%Y or (%y [and %C])) and (%b or %B or %m) and (%d or %e)
///	(%Y or (%y [and %C])) and (%j)
///	(%Y or (%y [and %C])) and (%U or %W) and (%a or %A or %u or %w)
///	(%G or (%g [and %C])) and (%V) and (%a or %A or %u or %w)
///	@endcode
///
///	The following combinations of format specifiers results into a valid time for a CDateTime object:
///
///	@code
///	[%R]
///	[%T]
///	[%X]
///	[(%H or %I or %k or %l) [and (%M) [and (%S) [and (%N) [and (%z or %Z)]]]]]
///	@endcode
///
///	@param[in]	avNext		The input iterator.
///	@param[in]	avEnd		The input end iterator.
///	@param[in]	arIosBase	The stream formatting information.
///	@param[out]	arIoState	The stream state.
///	@param[out]	arDateTime	The date and time retrieved from the input stream.
///	@param[in]	arFormat	The format of the date and time retrieved from the input stream.
///
///	@return	The input iterator.
///
CDateTimeGet::TInIt CDateTimeGet::mvGet
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base& arIosBase,
	std::ios_base::iostate& arIoState,
	CDateTime& arDateTime,
	const std::string& arFormat
) const
{
	return mvDoGet(avNext, avEnd, arIosBase, arIoState, arDateTime, arFormat);
}

///	@brief This function implements the behavior of the mvGet() function.
///
///	@param[in]	avNext		The input iterator.
///	@param[in]	avEnd		The input end iterator.
///	@param[in]	arIosBase	The stream formatting information.
///	@param[out]	arIoState	The stream state.
///	@param[out]	arDateTime	The date and time retrieved from the input stream.
///	@param[in]	arFormat	The format of the date and time retrieved from the input stream.
///
///	@return	The input iterator.
///
CDateTimeGet::TInIt CDateTimeGet::mvDoGet
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base& arIosBase,
	std::ios_base::iostate& arIoState,
	CDateTime& arDateTime,
	const std::string& arFormat
) const
{
	// There are some format specifiers which are simply aliases for a combination of other format
	// specifiers.  Replace these format specifiers with their equivalent format specifiers.
	string vFormat = smvExpandFormat(arFormat);

	// Initialize required variables.
	auto vCentury = -1;
	auto vYear = -1;
	auto vMonth = -1;
	auto vDay = -1;
	auto vDayOfYear = -1;
	auto vWeek = -1;
	auto vFirstDay = -1;
	auto vDayOfWeek = -1;
	auto vWeekBasedYear = -1;
	auto vWeekBasedWeek = -1;
	auto vHour = -1;
	auto vMinute = -1;
	auto vSecond = -1;
	auto vMillisecond = -1;
	auto vDst = -1;
	auto vHasAmPm = false;
	auto vIsPm = false;

	// Parse the date and time according to the current date and time format string.
	// Duplication of date and time parts is not allowed.
	string::size_type vFormatOffset {0};
	while (vFormatOffset < vFormat.length() && avNext != avEnd)
	{
		// Check for all recognized format specifiers.  If the characters from the input stream do not
		// match what we are parsing for then the input will be considered bad and the input iterator
		// will be pointing to the offending character.
		if (vFormat.compare(vFormatOffset, 2, "%%") == 0)
		{
			// Parse for % character.
			avNext = smvIgnoreChar(avNext, avEnd, arIoState, '%');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%a") == 0)
		{
			// Parse for abbreviated weekday name.
			avNext = smvGetWeekDayName(avNext, avEnd, arIosBase, arIoState, vDayOfWeek, true);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%A") == 0)
		{
			// Parse for full weekday name.
			avNext = smvGetWeekDayName(avNext, avEnd, arIosBase, arIoState, vDayOfWeek, false);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%b") == 0)
		{
			// Parse for abbreviated month name.
			avNext = smvGetMonthName(avNext, avEnd, arIosBase, arIoState, vMonth, true);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%B") == 0)
		{
			// Parse for full month name.
			avNext = smvGetMonthName(avNext, avEnd, arIosBase, arIoState, vMonth, false);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%C") == 0)
		{
			// Parse for 2-digit century number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vCentury, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%d") == 0)
		{
			// Parse for 2-digit day number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vDay, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%e") == 0)
		{
			// Parse for 2-digit day number blank padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vDay, 2, ' ');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%g") == 0)
		{
			// Parse for 2-digit week-based year without century zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vWeekBasedYear, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%G") == 0)
		{
			// Parse for 4-digit week-based year with century zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vWeekBasedYear, 4, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%H") == 0)
		{
			// Parse for 2-digit hour number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vHour, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%I") == 0)
		{
			// Parse for 2-digit hour number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vHour, 2, '0');
			vHour = vHour % 12;
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%j") == 0)
		{
			// Parse for 3-digit day of the year number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vDayOfYear, 3, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%k") == 0)
		{
			// Parse for 2-digit hour number blank padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vHour, 2, ' ');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%l") == 0)
		{
			// Parse for 2-digit hour number blank padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vHour, 2, ' ');
			vHour = vHour % 12;
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%m") == 0)
		{
			// Parse for 2-digit month number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vMonth, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%M") == 0)
		{
			// Parse for 2-digit minute number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vMinute, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%n") == 0)
		{
			// Parse for newline character.
			avNext = smvIgnoreChar(avNext, avEnd, arIoState, '\n');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%N") == 0)
		{
			// Parse for 3-digit millisecond number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vMillisecond, 3, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%p") == 0)
		{
			// Parse for the AM/PM indicator.
			avNext = smvGetAmPm(avNext, avEnd, arIosBase, arIoState, vHasAmPm, vIsPm);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%S") == 0)
		{
			// Parse for 2-digit second number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vSecond, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%t") == 0)
		{
			// Parse for horizontal tab character.
			avNext = smvIgnoreChar(avNext, avEnd, arIoState, '\t');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%u") == 0)
		{
			// Parse for 1-digit weekday number.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vDayOfWeek, 1, '0');
			vDayOfWeek = vDayOfWeek % CDate::DAY_PER_WEEK;
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%U") == 0)
		{
			// Parse for 2-digit week number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vWeek, 2, '0');
			vFirstDay = CDate::FIRST_DAY_SUN;
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%V") == 0)
		{
			// Parse for 2-digit week number of a week-based year zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vWeekBasedWeek, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%w") == 0)
		{
			// Parse for 1-digit weekday number.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vDayOfWeek, 1, '0');
			vDayOfWeek = vDayOfWeek % CDate::DAY_PER_WEEK;
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%W") == 0)
		{
			// Parse for 2-digit week number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vWeek, 2, '0');
			vFirstDay = CDate::FIRST_DAY_MON;
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%y") == 0)
		{
			// Parse for 2-digit year number without century zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vYear, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%Y") == 0)
		{
			// Parse for 4-digit year number with century zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vYear, 4, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%z") == 0)
		{
			// Parse for time zone offset.
			avNext = smvGetTzOffset(avNext, avEnd, arIosBase, arIoState, vDst);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%Z") == 0)
		{
			// Parse for time zone name.
			avNext = smvGetTzName(avNext, avEnd, arIoState, vDst);
			vFormatOffset += 2;
		}
		else
		{
			// Parse for normal character.
			avNext = smvIgnoreChar(avNext, avEnd, arIoState, vFormat[vFormatOffset]);
			++vFormatOffset;
		}

		// Check if the parsing is successful so far.
		if (arIoState != ios_base::goodbit)
		{
			// No, it isn't.
			return avNext;
		}
	}

	// If hour, minute, second and millisecond were not specified in the date and time format, set them
	// to their default values.
	if (vHour == -1)
	{
		vHour = 0;
	}
	if (vMinute == -1)
	{
		vMinute = 0;
	}
	if (vSecond == -1)
	{
		vSecond = 0;
	}
	if (vMillisecond == -1)
	{
		vMillisecond = 0;
	}
	if (vHasAmPm)
	{
		if (vIsPm)
		{
			vHour += 12;
		}
	}

	// If year greater than or equal to 0 and is less than 100, we need to add a century to it.
	if (vYear >= 0 && vYear < 100)
	{
		if (vCentury == -1)
		{
			CDate vDate;
			vYear += (vDate.mvGetDate().mvYear / 100) * 100;
		}
		else
		{
			vYear += vCentury * 100;
		}
	}

	// If week-based year greater than or equal to 0 and is less than 100, we need to add a century to it.
	if (vWeekBasedYear >= 0 && vWeekBasedYear < 100)
	{
		if (vCentury == -1)
		{
			CDate vDate;
			vWeekBasedYear += (vDate.mvGetDate().mvYear / 100) * 100;
		}
		else
		{
			vWeekBasedYear += vCentury * 100;
		}
	}

	// Setup the date to be returned.  It's bad if it generates an exception which we need to handle.
	if (vYear != -1 && vMonth != -1 && vDay != -1)
	{
		// Requires (%Y or (%y and optionally %C)) and (%b or %B or %m) and (%d or %e)
		// Optionally (%H or %I or %k or %l) and (%M) and (%S) and (%N) and (%z or %Z)
		arDateTime.mvSetDateTime(vYear, vMonth, vDay, vHour, vMinute, vSecond, vMillisecond);
	}
	else if (vYear != -1 && vDayOfYear != -1)
	{
		// Requires (%Y or (%y and optionally %C)) and (%j)
		// Optionally (%H or %I or %k or %l) and (%M) and (%S) and (%N) and (%z or %Z)
		arDateTime.mvSetDateTime(vYear, 1, 1, vHour, vMinute, vSecond, vMillisecond);
		arDateTime += CDateTimeDuration(vDayOfYear - 1, CDateTimeDuration::E_U_DAY);
	}
	else if (vYear != -1 && vWeek != -1 && vDayOfWeek != -1)
	{
		// Requires (%Y or (%y and optionally %C)) and (%U or %W) and (%a or %A or %u or %w)
		// Optionally (%H or %I or %k or %l) and (%M) and (%S) and (%N)
		// Determine the starting day of the year.
		CDate vDate1 {vYear, 1, 1};
		while (vDate1 % CDate::DAY_PER_WEEK != vFirstDay)
		{
			++vDate1;
		}
		auto vDate2 = vDate1.mvGetDate();
		arDateTime.mvSetDateTime
		(
			vDate2.mvYear,
			vDate2.mvMonth,
			vDate2.mvDay,
			vHour,
			vMinute,
			vSecond,
			vMillisecond
		);
		// Add the number of weeks.
		arDateTime += CDateTimeDuration(vWeek - 1, CDateTimeDuration::E_U_WEEK);
		// Add the number of days.
		arDateTime += CDateTimeDuration
		(
			(
				(vFirstDay == CDate::FIRST_DAY_SUN)
				?
				vDayOfWeek
				:
				((vDayOfWeek + CDate::DAY_PER_WEEK - 1) % CDate::DAY_PER_WEEK)
			),
			CDateTimeDuration::E_U_DAY
		);
	}
	else if (vWeekBasedYear != -1 && vWeekBasedWeek != -1 && vDayOfWeek != -1)
	{
		// Requires (%G or (%g and optionally %C)) and (%V) and (%a or %A or %u or %w)
		// Optionally (%H or %I or %k or %l) and (%M) and (%S) and (%N)
		// Determine the starting day of the year.
		CDate vDate1 {vWeekBasedYear, 1, 4};
		while (vDate1 % CDate::DAY_PER_WEEK != 0)
		{
			--vDate1;
		}
		auto vDate2 = vDate1.mvGetDate();
		arDateTime.mvSetDateTime
		(
			vDate2.mvYear,
			vDate2.mvMonth,
			vDate2.mvDay,
			vHour,
			vMinute,
			vSecond,
			vMillisecond
		);
		// Add the number of weeks.
		arDateTime += CDateTimeDuration(vWeekBasedWeek - 1, CDateTimeDuration::E_U_WEEK);
		// Add the number of days.
		arDateTime += CDateTimeDuration
		(
			((vDayOfWeek + CDate::DAY_PER_WEEK - 1) % CDate::DAY_PER_WEEK),
			CDateTimeDuration::E_U_DAY
		);
	}
	else
	{
		arIoState |= ios_base::failbit;
	}
	return avNext;
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
