//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include "lfc1/datetime/cdatetime.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief This function obtains the date and time format from a stream.
///
///	@param[in]	arIosBase	The stream containing the date and time format.
///
///	@return The date and time format.
///
std::string CDateTimeManip::smvGetFormat(std::ios_base& arIosBase)
{
	auto pFormat = reinterpret_cast<string*>(arIosBase.pword(CDateTime::smvGetIndex()));
	if (pFormat == nullptr || pFormat->empty())
	{
		return "%c";
	}
	else
	{
		return *pFormat;
	}
}

///	@brief This function obtains the error information relating to date and time format copying.
///
///	@param[in]	arIos	The stream containing the error information.
///
///	@return The error information relating to date and time format copying.
///
std::ios_base::iostate CDateTimeManip::smvCopyFmtErr(std::basic_ios<char>& arIos)
{
	auto vErr = static_cast<ios_base::iostate>(arIos.iword(CDateTime::smvGetIndex()));
	arIos.iword(CDateTime::smvGetIndex()) = 0;
	arIos.setstate(vErr | arIos.rdstate());
	return vErr;
}

///	@brief This function creates a CDateTimeManip object.
///
///	@param[in]	apFormat	The desired date and time format.
///
CDateTimeManip::CDateTimeManip(const char* apFormat)
:	CManipBase<const char*> {smvSetFormat, apFormat}
{
}

///	@brief This function destroys a CDateTimeManip object.
///
CDateTimeManip::~CDateTimeManip() noexcept
{
}

///	@brief This function stores the given date and time format in a stream.
///
///	@param[in]	arIosBase	The stream containing the date and time format.
///	@param[in]	apFormat	The date and time format.
///
///	@exception	std::bad_alloc	Indicates a memory allocation failure.
///
void CDateTimeManip::smvSetFormat(std::ios_base& arIosBase, const char* apFormat)
{
	// Obtain the current value of the pword for date and time format.
	string*& rPword = reinterpret_cast<string*&>(arIosBase.pword(CDateTime::smvGetIndex()));

	// Check if the pword is already set or not.
	if (rPword == nullptr)
	{
		// No it is not, register callback function.
		arIosBase.register_callback(smvCallback, CDateTime::smvGetIndex());
	}
	string* pNewFormat = new string(apFormat);
	string* pOldFormat = rPword;
	rPword = pNewFormat;
	delete pOldFormat;
}

///	@brief This function is the stream callback function.
///
///	@param[in]	avEvent		The stream event.
///	@param[in]	arIosBase	The stream.
///	@param[in]	avIndex		The pword index.
///
void CDateTimeManip::smvCallback(std::ios_base::event avEvent, std::ios_base& arIosBase, int avIndex) noexcept
{
	if (avEvent == ios_base::erase_event)
	{
		try
		{
			delete reinterpret_cast<string*>(arIosBase.pword(avIndex));
		}
		catch (...)
		{
			// Ignore the exception because the stream is about to be destroyed anyway.
		}
	}
	else if (avEvent == ios_base::copyfmt_event)
	{
		string*& rPword = reinterpret_cast<string*&>(arIosBase.pword(avIndex));
		if (rPword != nullptr)
		{
			try
			{
				rPword = new string(*rPword);
			}
			catch (bad_alloc&)
			{
				// Store the error in the stream's iword.  It is the users responsibility to fetch the
				// error stored in the stream's iword after invoking the stream's copyfmt() function.
				arIosBase.iword(avIndex) =
					static_cast<ios_base::iostate>(arIosBase.iword(avIndex)) | ios_base::badbit;
			}
		}
	}
}

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This function is the date and time manipulator.  It is named following the naming convention of
///		   the standard manipulators.
///
///	@param[in]	apFormat	The date and time format.  A null pointer or empty string will be considered to
///							mean use the default format.
///
///	@return The date and time manipulator helper.
///
CDateTimeManip setdatetimeformat(const char* apFormat)
{
	return CDateTimeManip {apFormat};
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
