//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include "lfc1/datetime/cdatetime.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief The facet ID.
///
std::locale::id CDateTimePut::id;

///	@brief This function creates a CDateTimePut object.
///
///	@param[in]	avRelease	Indicates who controls the lifetime of the facet. (0 means locale)
///
CDateTimePut::CDateTimePut(size_t avRelease)
:	locale::facet {avRelease}
{
}

///	@brief This function destroys a CDateTimePut object.
///
CDateTimePut::~CDateTimePut() noexcept
{
}

///	@brief This function outputs a date and time to a stream.
///
///	This function recognizes the following format specifiers which are identical to the date specific format
///	specifiers of the C language strftime() function except for %N:
///
///	@code
///	%% - A percent character.
///	%a - Abbreviated weekday name.
///	%A - Full weekday name.
///	%b - Abbreviated month name.
///	%B - Full month name.
///	%c - Default date and time format.
///	%C - Century [00-99].
///	%d - Day of the month zero padded [01-31].
///	%D - Equivalent to %m/%d/%y.
///	%e - Day of the month blank padded [ 1-31].
///	%F - Equivalent to %Y-%m-%d.
///	%g - Week-based year without century [00-99].	(See week-based year explanation below.)
///	%G - Week-based year with century [0000-9999].	(See week-based year explanation below.)
///	%h - Equivalent to %b.
///	%H - Hour in 24-hour format [00-23] zero padded.
///	%I - Hour in 12-hour format [01-12] zero padded.
///	%j - Day of the year [001-366].
///	%k - Hour in 24-hour format [00-23] blank padded.
///	%l - Hour in 12-hour format [01-12] blank padded.
///	%m - Month number [01-12].
///	%M - Minute [00-59].
///	%n - Newline character.
///	%N - Millisecond [000-999].
///	%p - AM/PM designation.
///	%r - Default 12-hour time format.
///	%R - Equivalent to %H:%M.
///	%S - Second [00-59].
///	%t - Horizontal tab character.
///	%T - Equivalent to %H:%M:%S.
///	%u - Day of the week [1-7], Monday is 1.
///	%U - Week of the year [00-53], the first Sunday starts week 1.
///	%V - Week number of a week-based year [01-53].	(See week-based year explanation below.)
///	%w - Day of the week [0-6], Sunday is 0.
///	%W - Week of the year [00-53], the first Monday starts week 1.
///	%x - Default date format.
///	%X - Default time format.
///	%y - Year without century [00-99].
///	%Y - Year with century [0000-9999].
///	%z - Time zone offset.
///	%Z - Time zone name.
///	@endcode
///
///	Any other sequence of characters not included in the above list will be taken literally.
///
///	For an explanation of the week-based year, see CDatePut.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	arDateTime	The date and time to be sent to the output stream.
///	@param[in]	arFormat	The format of the date and time to be sent to the output stream.
///
///	@return	The output iterator.
///
CDateTimePut::TOutIt CDateTimePut::mvPut
(
	TOutIt avOut,
	const CDateTime& arDateTime,
	const std::string& arFormat
) const
{
	return mvDoPut(avOut, arDateTime, arFormat);
}

///	@brief This function implements the behavior of the mvPut() function.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	arDateTime	The date and time to be sent to the output stream.
///	@param[in]	arFormat	The format of the date and time to be sent to the output stream.
///
///	@return	The output iterator.
///
CDateTimePut::TOutIt CDateTimePut::mvDoPut
(
	TOutIt avOut,
	const CDateTime& arDateTime,
	const std::string& arFormat
) const
{
	// Determine the equivalent year, month and day for the given CDate object.
	auto vDateTime = arDateTime.mvGetDateTime();
	CDate vDate {vDateTime.mvYear, vDateTime.mvMonth, vDateTime.mvDay};

	// There are some format specifiers which are simply aliases for a combination of other format
	// specifiers.  Replace these format specifiers with their equivalent format specifiers.
	auto vFormat = smvExpandFormat(arFormat);

	// Send output via the iterator.
	string::size_type vFormatOffset {0};
	while (vFormatOffset < vFormat.length())
	{
		// Check for all recognized format specifiers.
		if (vFormat.compare(vFormatOffset, 2, "%%") == 0)
		{
			// Output % character.
			avOut = '%';
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%a") == 0)
		{
			// Output abbreviated weekday name.
			avOut = smvPutWeekDayName(avOut, (vDate + 1) % CDate::DAY_PER_WEEK, true);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%A") == 0)
		{
			// Output full weekday name.
			avOut = smvPutWeekDayName(avOut, (vDate + 1) % CDate::DAY_PER_WEEK, false);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%b") == 0)
		{
			// Output abbreviated month name.
			avOut = smvPutMonthName(avOut, vDateTime.mvMonth, true);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%B") == 0)
		{
			// Output full month name.
			avOut = smvPutMonthName(avOut, vDateTime.mvMonth, false);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%C") == 0)
		{
			// Output 2-digit century number zero padded.
			avOut = smvPutNumber2Zero(avOut, vDateTime.mvYear / 100);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%d") == 0)
		{
			// Output 2-digit day number zero padded.
			avOut = smvPutNumber2Zero(avOut, vDateTime.mvDay);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%e") == 0)
		{
			// Output 2-digit day number blank padded.
			avOut = smvPutNumber2Blank(avOut, vDateTime.mvDay);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%g") == 0)
		{
			// Output 2-digit week-based year number without century zero padded.
			int vWeekBasedYear = vDateTime.mvYear + 1;
			smvCalcWeekBasedDetails(vDate, vWeekBasedYear, nullptr);
			avOut = smvPutNumber2Zero(avOut, vWeekBasedYear % 100);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%G") == 0)
		{
			// Output 4-digit week-based year number with century zero padded.
			int vWeekBasedYear = vDateTime.mvYear + 1;
			smvCalcWeekBasedDetails(vDate, vWeekBasedYear, nullptr);
			avOut = smvPutNumber4Zero(avOut, vWeekBasedYear);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%H") == 0)
		{
			// Output 2-digit 24-hour number zero padded.
			avOut = smvPutNumber2Zero(avOut, vDateTime.mvHour);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%I") == 0)
		{
			// Output 2-digit 12-hour number zero padded.
			int vHour = vDateTime.mvHour % 12;
			avOut = smvPutNumber2Zero(avOut, (vHour) ? vHour : 12);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%j") == 0)
		{
			// Output 3-digit day of the year number zero padded.
			CDate vStartYear {vDateTime.mvYear, 1, 1};
			avOut = smvPutNumber3Zero(avOut, vDate - vStartYear + 1);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%k") == 0)
		{
			// Output 2-digit 24-hour number blank padded.
			avOut = smvPutNumber2Blank(avOut, vDateTime.mvHour);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%l") == 0)
		{
			// Output 2-digit 12-hour number blank padded.
			int vHour = vDateTime.mvHour % 12;
			avOut = smvPutNumber2Blank(avOut, (vHour) ? vHour : 12);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%m") == 0)
		{
			// Output 2-digit month number zero padded.
			avOut = smvPutNumber2Zero(avOut, vDateTime.mvMonth);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%M") == 0)
		{
			// Output 2-digit minute number zero padded.
			avOut = smvPutNumber2Zero(avOut, vDateTime.mvMinute);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%n") == 0)
		{
			// Output newline character.
			avOut = '\n';
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%N") == 0)
		{
			// Output 3-digit millisecond number zero padded.
			avOut = smvPutNumber3Zero(avOut, vDateTime.mvMillisecond);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%p") == 0)
		{
			// Output A.M. or P.M. indicator.
			avOut = smvPutAmPm(avOut, vDateTime.mvHour);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%S") == 0)
		{
			// Output 2-digit second number zero padded.
			avOut = smvPutNumber2Zero(avOut, vDateTime.mvSecond);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%t") == 0)
		{
			// Output horizontal tab character.
			avOut = '\t';
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%u") == 0)
		{
			// Output 1-digit weekday number, Monday is 1.
			avOut = smvPutNumber1(avOut, (vDate + 0) % CDate::DAY_PER_WEEK + 1);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%U") == 0)
		{
			// Output 2-digit week number zero padded with weeks beginning on Sunday.
			int vWeekNo = smvCalcWeekNo(vDate, vDateTime.mvYear, CDate::FIRST_DAY_SUN);
			avOut = smvPutNumber2Zero(avOut, vWeekNo);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%V") == 0)
		{
			// Output 2-digit week number of a week-based year zero padded.
			int vWeekBasedYear = vDateTime.mvYear + 1;
			int vWeekBasedWeekNo;
			smvCalcWeekBasedDetails(vDate, vWeekBasedYear, &vWeekBasedWeekNo);
			avOut = smvPutNumber2Zero(avOut, vWeekBasedWeekNo);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%w") == 0)
		{
			// Output 1-digit weekday number, Sunday is 0.
			avOut = smvPutNumber1(avOut, (vDate + 1) % CDate::DAY_PER_WEEK);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%W") == 0)
		{
			// Output 2-digit week number zero padded with weeks beginning on Monday.
			int vWeekNo = smvCalcWeekNo(vDate, vDateTime.mvYear, CDate::FIRST_DAY_MON);
			avOut = smvPutNumber2Zero(avOut, vWeekNo);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%y") == 0)
		{
			// Output 2-digit year number without century zero padded.
			avOut = smvPutNumber2Zero(avOut, vDateTime.mvYear % 100);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%Y") == 0)
		{
			// Output 4-digit year number with century zero padded.
			avOut = smvPutNumber4Zero(avOut, vDateTime.mvYear);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%z") == 0)
		{
			// Output timezone offset.
			avOut = smvPutTzOffset(avOut, vDateTime.mvDst);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%Z") == 0)
		{
			// Output timezone name.
			avOut = smvPutTzName(avOut, vDateTime.mvDst);
			vFormatOffset += 2;
		}
		else
		{
			// Output normal character.
			avOut = vFormat[vFormatOffset];
			++vFormatOffset;
		}
	}
	return avOut;
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
