//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include "lfc1/datetime/cfacetbase.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief This function parses an input iterator for a specific character.  This function moves the input
///		   iterator forward one character if the expected character is found.
///
///	@param[in]	avNext		The input iterator.
///	@param[in]	avEnd		The input end iterator.
///	@param[out]	arIoState	The stream state.
///	@param[in]	avExpectedChar	The next character expected from the input iterator.
///
///	@retval	ios_base::goodbit	The expected character was found.
///	@retval	ios_base::failbit	The expected character was not found.
///
CFacetBase::TInIt CFacetBase::smvIgnoreChar
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base::iostate& arIoState,
	char avExpectedChar
) noexcept
{
	if (avNext == avEnd)
	{
		arIoState |= ios_base::failbit;
		return avNext;
	}

	if (*avNext == avExpectedChar)
	{
		++avNext;
	}
	else
	{
		arIoState |= ios_base::failbit;
	}
	return avNext;
}

///	@brief This function parses an input iterator for an integer value.
///
///	@param[in]	avNext		The input iterator.
///	@param[in]	avEnd		The input end iterator.
///	@param[in]	arIosBase	The stream formatting information.
///	@param[out]	arIoState	The stream state.
///	@param[out]	apNumber	The storage for the resulting integer.
///	@param[in]	avDigits	The maximum number of digits allowed for the integer.
///	@param[in]	avPadding	The padding character allowed for the integer value.
///
///	@retval	ios_base::goodbit	A valid integer was found.
///	@retval	ios_base::failbit	A valid integer was not found.
///
CFacetBase::TInIt CFacetBase::smvGetNumber
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base& arIosBase,
	std::ios_base::iostate& arIoState,
	int* apNumber,
	int avDigits,
	char avPadding
) noexcept
{
	// It's an error if we already have a number.
	if (apNumber != nullptr && *apNumber > -1)
	{
		arIoState |= ios_base::failbit;
		return avNext;
	}

	// Initialize the number.
	if (apNumber != nullptr)
	{
		*apNumber = 0;
	}

	// Parse for number.
	const auto& rCtype = use_facet<ctype<char>>(arIosBase.getloc());
	for (auto vI = 0; vI < avDigits; ++vI)
	{
		// Check if we have reached the end of input.
		if (avNext == avEnd)
		{
			arIoState |= ios_base::failbit;
			break;
		}

		// Check for digits and possible padding.
		if (vI < (avDigits - 1))
		{
			// Digits and padding character allowed before the last digit position.
			if (rCtype.is(ctype_base::digit, *avNext))
			{
				// We have found a digit.
				if (apNumber != nullptr)
				{
					*apNumber = *apNumber * 10 + *avNext - '0';
				}
			}
			else if (*avNext != avPadding)
			{
				// We have found a character other than a digit or padding character.
				arIoState |= ios_base::failbit;
				break;
			}
		}
		else
		{
			// Digits only allowed on the last digit position.
			if (rCtype.is(ctype_base::digit, *avNext))
			{
				// We have found a digit.
				if (apNumber != nullptr)
				{
					*apNumber = *apNumber * 10 + *avNext - '0';
				}
			}
			else
			{
				// We have found a character other than a digit.
				arIoState |= ios_base::failbit;
				break;
			}
		}

		// Move on to the next digit.
		++avNext;
	}
	return avNext;
}

///	@brief This function sends a 1-digit number to an output iterator.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	avNumber	The 1-digit number.
///
///	@return	The output iterator.
///
CFacetBase::TOutIt CFacetBase::smvPutNumber1(TOutIt avOut, int avNumber) noexcept
{
	avOut = static_cast<char>(avNumber + '0');
	return avOut;
}

///	@brief This function sends a 2-digit zero padded number to an output iterator.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	avNumber	The 2-digit number.
///
///	@return	The output iterator.
///
CFacetBase::TOutIt CFacetBase::smvPutNumber2Zero(TOutIt avOut, int avNumber) noexcept
{
	avOut = static_cast<char>(avNumber / 10 + '0');
	avOut = static_cast<char>(avNumber % 10 + '0');
	return avOut;
}

///	@brief This function sends a 2-digit blank padded number to an output iterator.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	avNumber	The 2-digit number.
///
///	@return	The output iterator.
///
CFacetBase::TOutIt CFacetBase::smvPutNumber2Blank(TOutIt avOut, int avNumber) noexcept
{
	if (avNumber >= 10)
	{
		avOut = static_cast<char>(avNumber / 10 + '0');
	}
	else
	{
		avOut = ' ';
	}
	avOut = static_cast<char>(avNumber % 10 + '0');
	return avOut;
}

///	@brief This function sends a 3-digit zero padded number to an output iterator.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	avNumber	The 3-digit number.
///
///	@return	The output iterator.
///
CFacetBase::TOutIt CFacetBase::smvPutNumber3Zero(TOutIt avOut, int avNumber) noexcept
{
	avOut = static_cast<char>((avNumber / 100) + '0');
	avOut = static_cast<char>((avNumber % 100) / 10 + '0');
	avOut = static_cast<char>((avNumber % 10) + '0');
	return avOut;
}

///	@brief This function sends a 4-digit zero padded number to an output iterator.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	avNumber	The 4-digit number.
///
///	@return	The output iterator.
///
CFacetBase::TOutIt CFacetBase::smvPutNumber4Zero(TOutIt avOut, int avNumber) noexcept
{
	auto vHundreds = avNumber / 100;
	avOut = static_cast<char>(vHundreds / 10 + '0');
	avOut = static_cast<char>(vHundreds % 10 + '0');
	auto vHundredths = avNumber % 100;
	avOut = static_cast<char>(vHundredths / 10 + '0');
	avOut = static_cast<char>(vHundredths % 10 + '0');
	return avOut;
}

///	@brief This function creates a default CFacetBase object.
///
CFacetBase::CFacetBase() noexcept
{
}

///	@brief This function destroys a CFacetBase object.
///
CFacetBase::~CFacetBase() noexcept
{
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
