//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#if defined(__LINUX__) || defined(__UNIX__)
#include <locale.h>
#include <langinfo.h>
#endif
#if defined(__WIN__)
#include <cstring>
#include <windows.h>
#endif
#include <sstream>
#include <stdexcept>

#include "lfc1/datetime/clanginfo.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

#if defined(__WIN__)

///	@brief This function converts a Windows date format string to a POSIX-like date format string.
///
///	@param[in]	apDateFormat	The Windows date format string.
///
///	@return	The POSIX-like date format string.
///
static std::string gvConvertDateFormat(const char* apDateFormat)
{
	string vDateFormat;
	auto vIsQuoted = false;
	for (auto vI = 0; apDateFormat[vI]; ++vI)
	{
		// Check if we are inside quoted characters.
		if (vIsQuoted)
		{
			// We are inside a quoted string so check if we have reached the end of the quoted string.
			if (apDateFormat[vI] == '\'')
			{
				// End of quoted string found.
				vIsQuoted = false;
			}
			else
			{
				// Copy all characters inside a quoted string.
				vDateFormat += apDateFormat[vI];
			}
		}
		else
		{
			// We are not inside a quoted string so check if we are about to begin a quoted string.
			if (apDateFormat[vI] == '\'')
			{
				// Beginning of quoted string found.
				vIsQuoted = true;
			}
			else
			{
				// Check for special characters.
				if (strncmp(apDateFormat + vI, "dddd", 4) == 0)
				{
					vDateFormat += "%A";
					vI += 3;
				}
				else if (strncmp(apDateFormat + vI, "ddd", 3) == 0)
				{
					vDateFormat += "%a";
					vI += 2;
				}
				else if (strncmp(apDateFormat + vI, "dd", 2) == 0)
				{
					vDateFormat += "%d";
					++vI;
				}
				else if (strncmp(apDateFormat + vI, "d", 1) == 0)
				{
					vDateFormat += "%e";
				}
				else if (strncmp(apDateFormat + vI, "MMMM", 4) == 0)
				{
					vDateFormat += "%B";
					vI += 3;
				}
				else if (strncmp(apDateFormat + vI, "MMM", 3) == 0)
				{
					vDateFormat += "%b";
					vI += 2;
				}
				else if (strncmp(apDateFormat + vI, "MM", 2) == 0)
				{
					vDateFormat += "%m";
					++vI;
				}
				else if (strncmp(apDateFormat + vI, "M", 1) == 0)
				{
					vDateFormat += "%m";
				}
				else if (strncmp(apDateFormat + vI, "yyyyy", 5) == 0)
				{
					vDateFormat += "%Y";
					vI += 4;
				}
				else if (strncmp(apDateFormat + vI, "yyyy", 4) == 0)
				{
					vDateFormat += "%Y";
					vI += 3;
				}
				else if (strncmp(apDateFormat + vI, "yy", 2) == 0)
				{
					vDateFormat += "%y";
					++vI;
				}
				else if (strncmp(apDateFormat + vI, "y", 1) == 0)
				{
					vDateFormat += "%y";
				}
				else
				{
					vDateFormat += apDateFormat[vI];
				}
			}
		}
	}
	return vDateFormat;
}

///	@brief This function converts a Windows time format string into a POSIX-like time format string.
///
///	@param[in]	apTimeFormat	The Windows time format string.
///
///	@return	The POSIX-like time format string.
///
static std::string gvConvertTimeFormat(const char* apTimeFormat)
{
	string vTimeFormat;
	auto vIsQuoted = false;
	for (auto vI = 0; apTimeFormat[vI]; ++vI)
	{
		// Check if we are inside quoted characters.
		if (vIsQuoted)
		{
			// We are inside a quoted string so check if we have reached the end of the quoted string.
			if (apTimeFormat[vI] == '\'')
			{
				// End of quoted string found.
				vIsQuoted = false;
			}
			else
			{
				// Copy all characters inside a quoted string.
				vTimeFormat += apTimeFormat[vI];
			}
		}
		else
		{
			// We are not inside a quoted string so check if we are about to begin a quoted string.
			if (apTimeFormat[vI] == '\'')
			{
				// Beginning of quoted string found.
				vIsQuoted = true;
			}
			else
			{
				// Check for special characters.
				if (strncmp(apTimeFormat + vI, "hh", 2) == 0)
				{
					vTimeFormat += "%I";
					++vI;
				}
				else if (strncmp(apTimeFormat + vI, "h", 1) == 0)
				{
					vTimeFormat += "%I";
				}
				else if (strncmp(apTimeFormat + vI, "HH", 2) == 0)
				{
					vTimeFormat += "%H";
					++vI;
				}
				else if (strncmp(apTimeFormat + vI, "H", 1) == 0)
				{
					vTimeFormat += "%H";
				}
				else if (strncmp(apTimeFormat + vI, "mm", 2) == 0)
				{
					vTimeFormat += "%M";
					++vI;
				}
				else if (strncmp(apTimeFormat + vI, "m", 1) == 0)
				{
					vTimeFormat += "%M";
				}
				else if (strncmp(apTimeFormat + vI, "ss", 2) == 0)
				{
					vTimeFormat += "%S";
					++vI;
				}
				else if (strncmp(apTimeFormat + vI, "s", 1) == 0)
				{
					vTimeFormat += "%S";
				}
				else if (strncmp(apTimeFormat + vI, "tt", 2) == 0)
				{
					vTimeFormat += "%p";
					++vI;
				}
				else
				{
					vTimeFormat += apTimeFormat[vI];
				}
			}
		}
	}
	return vTimeFormat;
}

#endif // __WIN__

#if defined(__LINUX__) || defined(__UNIX__)

///	@brief This function removes extensions to the POSIX-like date and time format strings.
///
///	@param[in]	arFormat	The date and time format string to be corrected.
///
///	@return	The corrected date and time format string.
///
static std::string& grFixFormat(std::string& arFormat) noexcept
{
/*
#if defined(__HPUX__)
	// Assumes that there will be no %% sequences in the format.
	const auto& rCtype = use_facet<ctype<char>>(locale());
	for (auto vI = 0U; vI < arFormat.length(); ++vI)
	{
		if (arFormat[vI] == '%')
		{
			// Removes non-alphabetic characters after a %.
			while (!rCtype.is(ctype_base::alpha, arFormat[vI + 1]))
			{
				arFormat.erase(vI + 1, 1);
			}
		}
	}
#endif
*/
	return arFormat;
}

#endif // __UNIX__ || __LINUX__

///	@brief This function obtains the date format.
///
///	@return	The date format.
///
std::string CLangInfo::smvGetDateFormat()
{
	// Obtain default date format.
#if defined(__LINUX__) || defined(__UNIX__)
	string vLocaleValue = nl_langinfo(D_FMT);
	return grFixFormat(vLocaleValue);
#endif
#if defined(__WIN__)
	char vLocaleValue[128];
	GetLocaleInfoA(LOCALE_USER_DEFAULT, LOCALE_SSHORTDATE, vLocaleValue, sizeof(vLocaleValue));
	return gvConvertDateFormat(vLocaleValue);
#endif
}

///	@brief This function obtains the 24-hour time format.
///
///	@return	The 24-hour time format.
///
std::string CLangInfo::smvGetTime24Format()
{
	// Obtain default time format.
#if defined(__LINUX__) || defined(__UNIX__)
	string vLocaleValue = nl_langinfo(T_FMT);
	return grFixFormat(vLocaleValue);
#endif
#if defined(__WIN__)
	char vLocaleValue[128];
	GetLocaleInfoA(LOCALE_USER_DEFAULT, LOCALE_STIMEFORMAT, vLocaleValue, sizeof(vLocaleValue));
	return gvConvertTimeFormat(vLocaleValue);
#endif
}

///	@brief This function obtains the 12-hour time format.
///
///	@return	The 12-hour time format.
///
std::string CLangInfo::smvGetTime12Format()
{
	// Obtain default 12-hour time format.
#if defined(__LINUX__) || defined(__UNIX__)
	string vLocaleValue = nl_langinfo(T_FMT_AMPM);
	return grFixFormat(vLocaleValue);
#endif
#if defined(__WIN__)
	char vLocaleValue[128];
	GetLocaleInfoA(LOCALE_USER_DEFAULT, LOCALE_STIMEFORMAT, vLocaleValue, sizeof(vLocaleValue));
	return gvConvertTimeFormat(vLocaleValue);
#endif
}

///	@brief This function obtains the date and time format.
///
///	@return	The date and time format.
///
std::string CLangInfo::smvGetDateTimeFormat()
{
	// Obtain default date and time format.
#if defined(__LINUX__) || defined(__UNIX__)
	string vLocaleValue = nl_langinfo(D_T_FMT);
	return grFixFormat(vLocaleValue);
#endif
#if defined(__WIN__)
	char vLocaleValue1[128];
	GetLocaleInfoA(LOCALE_USER_DEFAULT, LOCALE_SSHORTDATE, vLocaleValue1, sizeof(vLocaleValue1));
	char vLocaleValue2[128];
	GetLocaleInfoA(LOCALE_USER_DEFAULT, LOCALE_STIMEFORMAT, vLocaleValue2, sizeof(vLocaleValue2));
	return
		gvConvertDateFormat(vLocaleValue1) + " " +
		gvConvertTimeFormat(vLocaleValue2);
#endif
}

///	@brief This function obtains the long name of the specified week day.
///
///	@param[in]	avWeekDay	The number of the week day in the range 0 to 6 with 0 being Sunday.
///
///	@return	The long name of the specified week day.
///
///	@exception	std::out_of_range	Indicates the given week day is out-of-range.
///
std::string CLangInfo::smvGetLongDayName(int avWeekDay)
{
	// Validate day number.
	if (avWeekDay < 0 || avWeekDay > 6)
	{
		ostringstream vStream;
		vStream
			<< "Week day number 'avWeekDay' must be in the range 0 to 6 at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range(vStream.str());
	}

	// Obtain full weekday name.
#if defined(__LINUX__) || defined(__UNIX__)
	// DAY_1 is Sunday.
	return nl_langinfo(DAY_1 + avWeekDay);
#endif
#if defined(__WIN__)
	char vLocaleValue[128];
	GetLocaleInfoA
	(
		LOCALE_USER_DEFAULT,
		LOCALE_SDAYNAME1 + (avWeekDay + 6) % 7,
		vLocaleValue,
		sizeof(vLocaleValue)
	);
	return vLocaleValue;
#endif
}

///	@brief This function obtains the short name of the specified week day.
///
///	@param[in]	avWeekDay	The number of the week day in the range 0 to 6 with 0 being Sunday.
///
///	@return	The short name of the specified week day.
///
///	@exception	std::out_of_range	Indicates the given week day is out-of-range.
///
std::string CLangInfo::smvGetShortDayName(int avWeekDay)
{
	// Validate day number.
	if (avWeekDay < 0 || avWeekDay > 6)
	{
		ostringstream vStream;
		vStream
			<< "Week day number 'avWeekDay' must be in the range 0 to 6 at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range(vStream.str());
	}

	// Obtain abbreviated weekday name.
#if defined(__LINUX__) || defined(__UNIX__)
	// ABDAY_1 is Sunday.
	return nl_langinfo(ABDAY_1 + avWeekDay);
#endif
#if defined(__WIN__)
	// Obtain local short date format.
	char vLocaleValue[128];
	GetLocaleInfoA
	(
		LOCALE_USER_DEFAULT,
		LOCALE_SABBREVDAYNAME1 + (avWeekDay + 6) % 7,
		vLocaleValue,
		sizeof(vLocaleValue)
	);
	return vLocaleValue;
#endif
}

///	@brief This function obtains the long name of the specified month.
///
///	@param[in]	avMonth	The number of the month in the range 1 to 12 with 1 being January.
///
///	@return	The long name of the specified month.
///
///	@exception	std::out_of_range	Indicates the given month is out-of-range.
///
std::string CLangInfo::smvGetLongMonthName(int avMonth)
{
	// Validate month number.
	if (avMonth < 1 || avMonth > 12)
	{
		ostringstream vStream;
		vStream
			<< "Month number 'avMonth' must be in the range 1 to 12 at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range(vStream.str());
	}

	// Obtain full month name.
#if defined(__LINUX__) || defined(__UNIX__)
	return nl_langinfo(MON_1 + avMonth - 1);
#endif
#if defined(__WIN__)
	char vLocaleValue[128];
	GetLocaleInfoA
	(
		LOCALE_USER_DEFAULT,
		LOCALE_SMONTHNAME1 + avMonth - 1,
		vLocaleValue,
		sizeof(vLocaleValue)
	);
	return vLocaleValue;
#endif
}

///	@brief This function obtains the short name of the specified month.
///
///	@param[in]	avMonth	The number of the month in the range 1 to 12 with 1 being January.
///
///	@return	The short name of the specified month.
///
///	@exception	std::out_of_range	Indicates the given month is out-of-range.
///
std::string CLangInfo::smvGetShortMonthName(int avMonth)
{
	// Validate month number.
	if (avMonth < 1 || avMonth > 12)
	{
		ostringstream vStream;
		vStream
			<< "Month number 'avMonth' must be in the range 1 to 12 at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range(vStream.str());
	}

	// Obtain abbreviated month name.
#if defined(__LINUX__) || defined(__UNIX__)
	return nl_langinfo(ABMON_1 + avMonth - 1);
#endif
#if defined(__WIN__)
	// Obtain local short date format.
	char vLocaleValue[128];
	GetLocaleInfoA
	(
		LOCALE_USER_DEFAULT,
		LOCALE_SABBREVMONTHNAME1 + avMonth - 1,
		vLocaleValue,
		sizeof(vLocaleValue)
	);
	return vLocaleValue;
#endif
}

///	@brief This function obtains the string for AM.
///
///	@return	The string for AM.
///
std::string CLangInfo::smvGetAMString()
{
	// Obtain AM string.
#if defined(__LINUX__) || defined(__UNIX__)
	return nl_langinfo(AM_STR);
#endif
#if defined(__WIN__)
	char vLocaleValue[128];
	GetLocaleInfoA(LOCALE_USER_DEFAULT, LOCALE_S1159, vLocaleValue, sizeof(vLocaleValue));
	return vLocaleValue;
#endif
}

///	@brief This function obtains the string for PM.
///
///	@return	The string for PM.
///
std::string CLangInfo::smvGetPMString()
{
	// Obtain PM string.
#if defined(__LINUX__) || defined(__UNIX__)
	return nl_langinfo(PM_STR);
#endif
#if defined(__WIN__)
	char vLocaleValue[128];
	GetLocaleInfoA(LOCALE_USER_DEFAULT, LOCALE_S2359, vLocaleValue, sizeof(vLocaleValue));
	return vLocaleValue;
#endif
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
