//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include <algorithm>
#include <sstream>
#include <stdexcept>
#include <cstring>
#if defined(__LINUX__) || defined(__UNIX__)
#include <sys/types.h>
#include <sys/timeb.h>
#endif
#if defined(__WIN__)
#include <sys\types.h>
#include <sys\timeb.h>
#endif

#include "lfc1/datetime/ctime.hpp"
#include "lfc1/udtsup/input.hpp"
#include "lfc1/udtsup/output.hpp"

using namespace std;
using namespace lfc1::udtsup;

namespace lfc1 { namespace datetime {

///	@brief This function obtains an iword/pword index.
///
///	@return	The iword/pword index.
///
int CTime::smvGetIndex() noexcept
{
	call_once(smvFlag, smvAllocateIndex);
	return smvIndex;
}

///	@brief This function validates a time to be in the range 00:00:00.000 to 23:59:59.999.
///
///	@param[in]	avHour			The hour of a time.  (0 - 23)
///	@param[in]	avMinute		The minute of a time.  (0 - 59)
///	@param[in]	avSecond		The second of a time.  (0 - 59)
///	@param[in]	avMillisecond	The millisecond of a time.  (0 - 999)
///
///	@exception	std::out_of_range	Indicates either the hour, minute, second or millisecond are not within
///									valid ranges.
///
void CTime::smvValidateTime(int avHour, int avMinute, int avSecond, int avMillisecond)
{
	// Check the range for the hour.
	if (avHour < MIN_HR || avHour > MAX_HR)
	{
		ostringstream vStream;
		vStream << "Hour 'avHour' must be in the range 0 to 23 at (" << __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}

	// Check the range for the minute.
	if (avMinute < MIN_MIN || avMinute > MAX_MIN)
	{
		ostringstream vStream;
		vStream
			<< "Minute 'avMinute' must be in the range 0 to 59 at (" << __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}

	// Check the range for the second.
	if (avSecond < MIN_SEC || avSecond > MAX_SEC)
	{
		ostringstream vStream;
		vStream
			<< "Second 'avSecond' must be in the range 0 to 59 at (" << __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}

	// Check the range for the millisecond.
	if (avMillisecond < MIN_MSEC || avMillisecond > MAX_MSEC)
	{
		ostringstream vStream;
		vStream
			<< "Millisecond 'avMillisecond' must be in the range 0 to 999 at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
}

///	@brief This function creates a default CTime object based on the current system time.
///
CTime::CTime() noexcept
:	mvTimeNo {0}
{
	mvSetTime();
}

///	@brief This function creates a CTime object using the given Julian day number.
///
///	@param[in]	avTimeNo	The milliseconds since midnight number to assign to the CTime object.
///
///	@exception	std::out_of_range	Indicates that the given milliseconds since midnight number is not within
///									the valid range.
///
CTime::CTime(int avTimeNo)
:	mvTimeNo {avTimeNo}
{
	if (avTimeNo < MIN_TIMENO || avTimeNo > MAX_TIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Julian day number 'avTimeNo' must be within the range " << MIN_TIMENO << " to " << MAX_TIMENO
			<< " at (" << __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
}

///	@brief This function creates a CTime object using the given hour, minute, second and millisecond.
///
///	@param[in]	avHour			The given hour.
///	@param[in]	avMinute		The given minute.
///	@param[in]	avSecond		The given second.
///	@param[in]	avMillisecond	The given millisecond.
///
CTime::CTime(int avHour, int avMinute, int avSecond, int avMillisecond)
:	mvTimeNo {0}
{
	mvSetTime(avHour, avMinute, avSecond, avMillisecond);
}

///	@brief This function copy constructs a CTime object.
///
///	@param[in]	arRHS	The CTime object to be copied.
///
CTime::CTime(const CTime& arRHS) noexcept
:	mvTimeNo {arRHS.mvTimeNo}
{
}

///	@brief This function move constructs a CTime object.
///
///	@param[in]	arRHS	The CTime object to be moved.
///
CTime::CTime(CTime&& arRHS) noexcept
:	mvTimeNo {arRHS.mvTimeNo}
{
}

///	@brief This function destroys a CTime object.
///
CTime::~CTime() noexcept
{
}

///	@brief This function assigns the given Julian day number to the CTime object.
///
///	@param[in]	avTimeNo	The milliseconds since mignight number to assign to the CTime object.
///
///	@return	The CTime object assigned to.
///
CTime& CTime::operator=(int avTimeNo)
{
	CTime vCopy {avTimeNo};
	swap(vCopy);
	return *this;
}

///	@brief This function assigns a CTime object to another CTime object.
///
///	@param[in]	arRHS	The CTime object to assign to another CTime object.
///
///	@return	The CTime object assigned to.
///
CTime& CTime::operator=(const CTime& arRHS) noexcept
{
	mvTimeNo = arRHS.mvTimeNo;
	return *this;
}

///	@brief This function moves a CTime object to another CTime object.
///
///	@param[in]	arRHS	The CTime object to move to another CTime object.
///
///	@return	The CTime object assigned to.
///
CTime& CTime::operator=(CTime&& arRHS) noexcept
{
	mvTimeNo = arRHS.mvTimeNo;
	return *this;
}

///	@brief This function obtains the equivalent milliseconds since midnight number of a CTime object.
///
///	@return The equivalent milliseconds since midnight number of a CTime object.
///
CTime::operator int () const noexcept
{
	return mvTimeNo;
}

///	@brief This function assigns the current system time to a CTime object.
///
void CTime::mvSetTime() noexcept
{
	// Obtain the current system time with millisecond precision.
#if defined(__LINUX__) || defined(__UNIX__)
	timeb vNow;
	ftime(&vNow);
#endif
#if defined(__WIN__)
	_timeb vNow;
	_ftime_s(&vNow);
#endif
	auto vSystemTime = vNow.time;
	auto vMillisecond = static_cast<int>(vNow.millitm);

	// Convert the current system time to a structure.
	tm vTm;
	memset(&vTm, 0, sizeof(vTm));
#if defined(__LINUX__) || defined(__UNIX__)
	localtime_r(&vSystemTime, &vTm);
#endif
#if defined(__WIN__)
	localtime_s(&vTm, &vSystemTime);
#endif
	mvSetTime(vTm.tm_hour, vTm.tm_min, vTm.tm_sec, vMillisecond);
}

///	@brief This function sets a CTime object based on the given hour, minute, second and millisecond.
///
///	@param[in]	avHour			The given hour.
///	@param[in]	avMinute		The given minute.
///	@param[in]	avSecond		The given second.
///	@param[in]	avMillisecond	The given millisecond.
///
void CTime::mvSetTime(int avHour, int avMinute, int avSecond, int avMillisecond)
{
	// Validate the hour, minute, second and millisecond.
	smvValidateTime(avHour, avMinute, avSecond, avMillisecond);

	// Convert the hour, minute, second and millisecond to number of milliseconds since midnight.
	mvTimeNo = avHour;
	mvTimeNo = mvTimeNo * MIN_PER_HR + avMinute;
	mvTimeNo = mvTimeNo * SEC_PER_MIN + avSecond;
	mvTimeNo = mvTimeNo * MSEC_PER_SEC + avMillisecond;
}

///	@brief This function obtains the equivalent hour, minute, second and millisecond of the CTime object.
///
///	@return The equivalent hour, minute, second and millisecond of the CTime object.
///
CTime::STime CTime::mvGetTime() const noexcept
{
	// Convert the number of milliseconds since midnight to hour, minute, second and milliseconds.
	STime vTime;
	int vTimeNo = mvTimeNo;
	vTime.mvMillisecond = vTimeNo % MSEC_PER_SEC;
	vTimeNo /= MSEC_PER_SEC;
	vTime.mvSecond = vTimeNo % SEC_PER_MIN;
	vTimeNo /= SEC_PER_MIN;
	vTime.mvMinute = vTimeNo % MIN_PER_HR;
	vTime.mvHour = vTimeNo / MIN_PER_HR;

	// Return the result.
	return vTime;
}

///	@brief This function swaps a CTime object with another CTime object.
///
///	@param[in,out]	arRHS	The CTime object to swap with.
///
void CTime::swap(CTime& arRHS) noexcept
{
	using std::swap;
	swap(mvTimeNo, arRHS.mvTimeNo);
}

///	@brief This function pre-increments the CTime object by one millisecond.
///
///	@return	The CTime object.
///
///	@exception	std::out_of_range	Indicates incrementing the time will make the time go past the maximum
///									supported time.
///
CTime& CTime::operator++()
{
	if ((mvTimeNo + 1) > MAX_TIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Milliseconds since midnight number 'mvTimeNo' exceeds maximum limit " << MAX_TIMENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
	++mvTimeNo;
	return *this;
}

///	@brief This function post-increments the CTime object by one millisecond.
///
///	@return	The CTime object before the increment.
///
CTime CTime::operator++(int)
{
	CTime vTime = *this;
	++(*this);
	return vTime;
}

///	@brief This function pre-decrements the CTime object by one millisecond.
///
///	@return	The CTime object.
///
///	@exception	std::out_of_range	Indicates incrementing the time will make the time go below the minimum
///									supported time.
///
CTime& CTime::operator--()
{
	if ((mvTimeNo - 1) < MIN_TIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Milliseconds since midnight number 'mvTimeNo' exceeds minimum limit " << MIN_TIMENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}
	--mvTimeNo;
	return *this;
}

///	@brief This function post-decrements the CTime object by one millisecond.
///
///	@return	The CTime object before the decrement.
///
CTime CTime::operator--(int)
{
	CTime vTime = *this;
	--(*this);
	return vTime;
}

///	@brief This function increments the CTime object by the given duration.
///
///	@param[in]	arTimeDuration	The duration to add to the time.
///
///	@return	The CTime object.
///
///	@exception	std::out_of_range	Indicates incrementing the time will make the time go past the maximum
///									supported time.
///
CTime& CTime::operator+=(const CTimeDuration& arTimeDuration)
{
	auto vDuration = arTimeDuration.mvGetDuration();

	// Convert the duration according to its unit of measure.
	switch (arTimeDuration.mvGetUnit())
	{
	case CTimeDuration::E_U_HOUR:
		vDuration *= MSEC_PER_HR;
		break;

	case CTimeDuration::E_U_MINUTE:
		vDuration *= MSEC_PER_MIN;
		break;

	case CTimeDuration::E_U_SECOND:
		vDuration *= MSEC_PER_SEC;
		break;

	case CTimeDuration::E_U_MILLISECOND:
		vDuration *= 1;
		break;

	default:
		break;
	}

	// Check for overflow.
	if ((mvTimeNo + vDuration) > MAX_TIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Milliseconds since midnight number 'mvTimeNo' exceeds maximum limit " << MAX_TIMENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}

	// Add the duration to the CTime object.
	mvTimeNo += vDuration;

	// Return the incremented value.
	return *this;
}

///	@brief This function decrements the CTime object by the given duration.
///
///	@param[in]	arTimeDuration	The duration to subtract from the time.
///
///	@return	The CTime object.
///
///	@exception	std::out_of_range	Indicates incrementing the time will make the time go below the minimum
///									supported time.
///
CTime& CTime::operator-=(const CTimeDuration& arTimeDuration)
{
	auto vDuration = arTimeDuration.mvGetDuration();

	// Convert the duration according to its unit of measure.
	switch (arTimeDuration.mvGetUnit())
	{
	case CTimeDuration::E_U_HOUR:
		vDuration *= MSEC_PER_HR;
		break;

	case CTimeDuration::E_U_MINUTE:
		vDuration *= MSEC_PER_MIN;
		break;

	case CTimeDuration::E_U_SECOND:
		vDuration *= MSEC_PER_SEC;
		break;

	case CTimeDuration::E_U_MILLISECOND:
		vDuration *= 1;
		break;

	default:
		break;
	}

	// Check for underflow.
	if ((mvTimeNo - vDuration) < MIN_TIMENO)
	{
		ostringstream vStream;
		vStream
			<< "Milliseconds since midnight number 'mvTimeNo' exceeds minimum limit " << MIN_TIMENO << " at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw out_of_range {vStream.str()};
	}

	// Subtract the duration from the CTime object.
	mvTimeNo -= vDuration;

	// Return the decremented value.
	return *this;
}

///	@brief This function receives a CTime object from the given input stream.
///
///	@param[in]	arStream	The source stream.
///
///	@return	The resulting stream state.
///
std::ios_base::iostate CTime::mvInput(std::istream& arStream)
{
	// Obtain the time format.
	auto vFormat = CTimeManip::smvGetFormat(arStream);

	// Initialize a local stream state.
	auto vIoState = ios_base::goodbit;

	// Get input from stream.
	istreambuf_iterator<char> vEnd;
	use_facet<CTimeGet>(arStream.getloc()).mvGet(arStream, vEnd, arStream, vIoState, *this, vFormat);
	return vIoState;
}

///	@brief This function sends a CTime object to the given output stream.
///
///	@param[in]	arStream	The destination stream.
///
///	@return	The resulting stream state.
///
std::ios_base::iostate CTime::mvOutput(std::ostream& arStream) const
{
	// Obtain the time format.
	auto vFormat = CTimeManip::smvGetFormat(arStream);

	// Initialize a local stream state.
	auto vIoState = ios_base::goodbit;

	// Create a string buffer.
	ostringstream vBuffer;

	// Send output to string buffer.
	if (use_facet<CTimePut>(arStream.getloc()).mvPut(vBuffer, *this, vFormat).failed())
	{
		vIoState = ios_base::badbit;
	}

	// Field width adjustment.
	if (vIoState == ios_base::goodbit)
	{
		// Save the contents of the string buffer to a string.
		auto vOutput = vBuffer.str();

		// Determine if we have to do any padding.
		auto vPadding = static_cast<int>(arStream.width() - vOutput.length());
		ostreambuf_iterator<char> vSink {arStream};
		if (vPadding <= 0)
		{
			copy(vOutput.begin(), vOutput.end(), vSink);
		}
		else if (arStream.flags() & ios_base::left)
		{
			copy(vOutput.begin(), vOutput.end(), vSink);
			fill_n(vSink, vPadding, arStream.fill());
		}
		else
		{
			fill_n(vSink, vPadding, arStream.fill());
			copy(vOutput.begin(), vOutput.end(), vSink);
		}
		if (vSink.failed())
		{
			vIoState = ios_base::failbit;
		}
	}
	return vIoState;
}

///	@brief This function allocates an iword/pword index.
///
void CTime::smvAllocateIndex() noexcept
{
	smvIndex = ios_base::xalloc();
}

int CTime::smvIndex;
std::once_flag CTime::smvFlag;

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/ctime.hpp
///	@brief This function overloads std::swap for CTime objects.
///
///	@param[in,out]	arLHS	The first object to swap.
///	@param[in,out]	arRHS	The second object to swap.
///
void swap(CTime& arLHS, CTime& arRHS) noexcept
{
	arLHS.swap(arRHS);
}

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/ctime.hpp
///	@brief This function receives a CTime object from an input stream.
///
///	@param[in]	arStream	The source input stream.
///	@param[out]	arTime		The received CTime object.
///
///	@return	The source input stream.
///
std::istream& operator>>(std::istream& arStream, CTime& arTime)
{
	return gvInputGuarded<CTime>(arStream, arTime);
}

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/ctime.hpp
///	@brief This function sends a CTime object to an output stream.
///
///	@param[in]	arStream	The destination output stream.
///	@param[in]	arTime		The CTime object to send.
///
///	@return	The destination output stream.
///
std::ostream& operator<<(std::ostream& arStream, const CTime& arTime)
{
	return gvOutputGuarded<CTime>(arStream, arTime);
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
