//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include "lfc1/datetime/ctime.hpp"
#include "lfc1/datetime/clanginfo.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief This function expands the time format by replacing format specifiers with their equivalent format
///		   specifiers.
///
///	@param[in]	avFormat	The time format to be expanded.
///
///	@return	The expanded format string.
///
std::string CTimeBase::smvExpandFormat(std::string avFormat)
{
	// Take a copy of the time format.
	string::size_type vPos;
	for (;;)
	{
		if ((vPos = avFormat.find("%r")) != string::npos)
		{
			// Replace %r if it exists in the time format.
			avFormat.replace(vPos, 2, CLangInfo::smvGetTime12Format());
		}
		else if ((vPos = avFormat.find("%R")) != string::npos)
		{
			// Replace %R if it exists in the time format.
			avFormat.replace(vPos, 2, "%H:%M");
		}
		else if ((vPos = avFormat.find("%T")) != string::npos)
		{
			// Replace %T if it exists in the time format.
			avFormat.replace(vPos, 2, "%H:%M:%S");
		}
		else if ((vPos = avFormat.find("%X")) != string::npos)
		{
			// Replace %X if it exists in the time format.
			avFormat.replace(vPos, 2, CLangInfo::smvGetTime24Format());
		}
		else
		{
			break;
		}
	}

	// Return the expanded time format.
	return avFormat;
}

///	@brief This function parses an input iterator for an AM/PM indicator.
///
///	@param[in]	avNext		The input iterator.
///	@param[in]	avEnd		The input end iterator.
///	@param[in]	arIosBase	The stream formatting information.
///	@param[out]	arIoState	The stream state.
///	@param[out]	arHasAmPm	The storage for the resulting has AM/PM indicator (true) or not (false).
///	@param[out]	arIsPm		The storage for the resulting PM (true) or AM (false) indicator.
///
///	@return	The input iterator.
///
CTimeBase::TInIt CTimeBase::smvGetAmPm
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base& arIosBase,
	std::ios_base::iostate& arIoState,
	bool& arHasAmPm,
	bool& arIsPm
)
{
	// It's an error if an AM/PM indicator already exists.
	if (arHasAmPm)
	{
		arIoState |= ios_base::failbit;
		return avNext;
	}

	// Parse for AM/PM indicator.
	const auto& rCtype = use_facet<ctype<char>>(arIosBase.getloc());
	string vAmPm;
	while (avNext != avEnd && rCtype.is(ctype_base::alpha, *avNext))
	{
		vAmPm += *avNext;
		++avNext;
	}

	// Find matching AM/PM strings.
	if (vAmPm == CLangInfo::smvGetAMString())
	{
		// AM string found.
		arHasAmPm = true;
		arIsPm = false;
	}
	else if (vAmPm == CLangInfo::smvGetPMString())
	{
		// PM string found.
		arHasAmPm = true;
		arIsPm = true;
	}
	else
	{
		// Am or PM string not found.
		arIoState |= ios_base::failbit;
	}
	return avNext;
}

///	@brief This function sends AM or PM to an output iterator based on the given hour.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	avHour	The given hour.
///
///	@return	The output iterator.
///
CTimeBase::TOutIt CTimeBase::smvPutAmPm(TOutIt avOut, int avHour)
{
	string vAmPmString;
	if (avHour < 12)
	{
		vAmPmString = CLangInfo::smvGetAMString();
	}
	else
	{
		vAmPmString = CLangInfo::smvGetPMString();
	}
	copy(vAmPmString.begin(), vAmPmString.end(), avOut);
	return avOut;
}

///	@brief This function creates a default CTimeBase object.
///
CTimeBase::CTimeBase() noexcept
{
}

///	@brief This function destroys a CTimeBase object.
///
CTimeBase::~CTimeBase() noexcept
{
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
