//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include <string>

#include "lfc1/datetime/ctime.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief The facet ID.
///
std::locale::id CTimeGet::id;

///	@brief This function creates a CTimeGet object.
///
///	@param[in]	avRelease	Indicates who controls the lifetime of the facet. (0 means locale)
///
CTimeGet::CTimeGet(size_t avRelease)
:	locale::facet {avRelease}
{
}

///	@brief This function destroys a CTimeGet object.
///
CTimeGet::~CTimeGet() noexcept
{
}

///	@brief This function inputs a time from a stream.  Duplication of time parts is not allowed.
///
///	This function recognizes the following format specifiers which are identical to the time specific format
///	specifiers of the C language strftime() function except for %N:
///
///	@code
///	%% - A percent character.
///	%H - Hour in 24-hour format [00-23] zero padded.
///	%I - Hour in 12-hour format [01-12] zero padded.
///	%k - Hour in 24-hour format [00-23] blank padded.
///	%l - Hour in 12-hour format [01-12] blank padded.
///	%M - Minute [00-59].
///	%n - Newline character.
///	%N - Millisecond [000-999].
///	%p - AM/PM designation.
///	%r - Default 12-hour time format.
///	%R - Equivalent to %H:%M.
///	%S - Second [00-59].
///	%t - Horizontal tab character.
///	%T - Equivalent to %H:%M:%S.
///	%X - Default time format.
///	@endcode
///
///	Any other sequence of characters not included in the above list will be taken literally.
///
///	The following combinations of format specifiers results into a valid CTime object:
///
///	@code
///	%R
///	%T
///	%X
///	(%H or %I or %k or %l) [and (%M) [and (%S) [and (%N)]]]
///	@endcode
///
///	@param[in]	avNext		The input stream iterator.
///	@param[in]	avEnd		The input stream end iterator.
///	@param[in]	arIosBase	The stream formatting information.
///	@param[out]	arIoState	The stream state.  Set to ios_base::failbit if parsing failed.
///	@param[out]	arTime		The time retrieved from the input stream.
///	@param[in]	arFormat	The format of the time retrieved from the input stream.
///
///	@return	The input iterator.
///
CTimeGet::TInIt CTimeGet::mvGet
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base& arIosBase,
	std::ios_base::iostate& arIoState,
	CTime& arTime,
	const std::string& arFormat
) const
{
	return mvDoGet(avNext, avEnd, arIosBase, arIoState, arTime, arFormat);
}

///	@brief This function implements the behavior of the mvGet() function.
///
///	@param[in]	avNext		The input iterator.
///	@param[in]	avEnd		The input end iterator.
///	@param[in]	arIosBase	The stream formatting information.
///	@param[out]	arIoState	The stream state.
///	@param[out]	arTime		The time retrieved from the input stream.
///	@param[in]	arFormat	The format of the time retrieved from the input stream.
///
///	@return	The input iterator.
///
CTimeGet::TInIt CTimeGet::mvDoGet
(
	TInIt avNext,
	TInIt avEnd,
	std::ios_base& arIosBase,
	std::ios_base::iostate& arIoState,
	CTime& arTime,
	const std::string& arFormat
) const
{
	// There are some format specifiers which are simply aliases for a combination of other format
	// specifiers.  Replace these format specifiers with their equivalent format specifiers.
	auto vFormat = smvExpandFormat(arFormat);

	// Initialize required variables.
	auto vHour = -1;
	auto vMinute = -1;
	auto vSecond = -1;
	auto vMillisecond = -1;
	auto vHasAmPm = false;
	auto vIsPm = false;

	// Parse the time according to the current time format string.
	// Duplication of time parts is not allowed.
	string::size_type vFormatOffset {0};
	while (vFormatOffset < vFormat.length() && avNext != avEnd)
	{
		// Check for all recognized format specifiers.  If the characters from the input stream do not
		// match what we are parsing for then the input will be considered bad and the input iterator
		// will be pointing to the offending character.
		if (vFormat.compare(vFormatOffset, 2, "%%") == 0)
		{
			// Parse for % character.
			avNext = smvIgnoreChar(avNext, avEnd, arIoState, '%');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%H") == 0)
		{
			// Parse for 2-digit hour number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vHour, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%I") == 0)
		{
			// Parse for 2-digit hour number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vHour, 2, '0');
			vHour = vHour % 12;
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%k") == 0)
		{
			// Parse for 2-digit hour number blank padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vHour, 2, ' ');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%l") == 0)
		{
			// Parse for 2-digit hour number blank padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vHour, 2, ' ');
			vHour = vHour % 12;
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%M") == 0)
		{
			// Parse for 2-digit minute number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vMinute, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%n") == 0)
		{
			// Parse for newline character.
			avNext = smvIgnoreChar(avNext, avEnd, arIoState, '\n');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%N") == 0)
		{
			// Parse for 3-digit millisecond number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vMillisecond, 3, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%p") == 0)
		{
			// Parse for the AM/PM indicator.
			avNext = smvGetAmPm(avNext, avEnd, arIosBase, arIoState, vHasAmPm, vIsPm);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%S") == 0)
		{
			// Parse for 2-digit second number zero padded.
			avNext = smvGetNumber(avNext, avEnd, arIosBase, arIoState, &vSecond, 2, '0');
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%t") == 0)
		{
			// Parse for horizontal tab character.
			avNext = smvIgnoreChar(avNext, avEnd, arIoState, '\t');
			vFormatOffset += 2;
		}
		else
		{
			// Parse for normal character.
			avNext = smvIgnoreChar(avNext, avEnd, arIoState, vFormat[vFormatOffset]);
			++vFormatOffset;
		}

		// Check if the parsing is successful so far.
		if (arIoState != ios_base::goodbit)
		{
			return avNext;
		}
	}

	// If minute, second and millisecond were not specified in the time format, set them to their
	// default values.
	if (vMinute == -1)
	{
		vMinute = 0;
	}
	if (vSecond == -1)
	{
		vSecond = 0;
	}
	if (vMillisecond == -1)
	{
		vMillisecond = 0;
	}
	if (vHasAmPm)
	{
		if (vIsPm)
		{
			vHour += 12;
		}
	}

	// Setup the time to be returned.  It's bad if it generates an exception which we need to handle.
	arTime.mvSetTime(vHour, vMinute, vSecond, vMillisecond);
	return avNext;
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
