//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include <string>

#include "lfc1/datetime/ctime.hpp"

using namespace std;

namespace lfc1 { namespace datetime {

///	@brief The facet ID.
///
std::locale::id CTimePut::id;

///	@brief This function creates a CTimePut object.
///
///	@param[in]	avRelease	Indicates who controls the lifetime of the facet. (0 means locale)
///
CTimePut::CTimePut(size_t avRelease)
:	locale::facet {avRelease}
{
}

///	@brief This function destroys a CTimePut object.
///
CTimePut::~CTimePut() noexcept
{
}

///	@brief This function outputs a time to a stream.
///
///	This function recognizes the following format specifiers which are identical to the time specific format
///	specifiers of the C language strftime() function except for %N:
///
///	@code
///	%% - A percent character.
///	%H - Hour in 24-hour format [00-23] zero padded.
///	%I - Hour in 12-hour format [01-12] zero padded.
///	%k - Hour in 24-hour format [00-23] blank padded.
///	%l - Hour in 12-hour format [01-12] blank padded.
///	%M - Minute [00-59].
///	%n - Newline character.
///	%N - Millisecond [000-999].
///	%p - AM/PM designation.
///	%r - Default 12-hour time format.
///	%R - Equivalent to %H:%M.
///	%S - Second [00-59].
///	%t - Horizontal tab character.
///	%T - Equivalent to %H:%M:%S.
///	%X - Default time format.
///	@endcode
///
///	Any other sequence of characters not included in the above list will be taken literally.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	arTime		The time to be sent to the output stream.
///	@param[in]	arFormat	The format of the time to be sent to the output stream.
///
///	@return	The output iterator.
///
CTimePut::TOutIt CTimePut::mvPut(TOutIt avOut, const CTime& arTime, const std::string& arFormat) const
{
	return mvDoPut(avOut, arTime, arFormat);
}

///	@brief This function implements the behavior of the mvPut() function.
///
///	@param[in]	avOut		The output iterator.
///	@param[in]	arTime		The time to be sent to the output stream.
///	@param[in]	arFormat	The format of the time to be sent to the output stream.
///
///	@return	The output iterator.
///
CTimePut::TOutIt CTimePut::mvDoPut(TOutIt avOut, const CTime& arTime, const std::string& arFormat) const
{
	// Determine the equivalent hour, minute, second and millisecond for the given CTime object.
	auto vTime = arTime.mvGetTime();

	// There are some format specifiers which are simply aliases for a combination of other format
	// specifiers.  Replace these format specifiers with their equivalent format specifiers.
	auto vFormat = smvExpandFormat(arFormat);

	// Build the string to be sent to the output stream according to the current time format string.
	string::size_type vFormatOffset {0};
	while (vFormatOffset < vFormat.length())
	{
		// Check for all recognized format specifiers.
		if (vFormat.compare(vFormatOffset, 2, "%%") == 0)
		{
			// Output % character.
			avOut = '%';
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%H") == 0)
		{
			// Output 2-digit 24-hour number zero padded.
			avOut = smvPutNumber2Zero(avOut, vTime.mvHour);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%I") == 0)
		{
			// Output 2-digit 12-hour number zero padded.
			int vHour = vTime.mvHour % 12;
			avOut = smvPutNumber2Zero(avOut, (vHour) ? vHour : 12);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%k") == 0)
		{
			// Output 2-digit 24-hour number blank padded.
			avOut = smvPutNumber2Blank(avOut, vTime.mvHour);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%l") == 0)
		{
			// Output 2-digit 12-hour number blank padded.
			int vHour = vTime.mvHour % 12;
			avOut = smvPutNumber2Blank(avOut, (vHour) ? vHour : 12);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%M") == 0)
		{
			// Output 2-digit minute number zero padded.
			avOut = smvPutNumber2Zero(avOut, vTime.mvMinute);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%n") == 0)
		{
			// Output newline character.
			avOut = L'\n';
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%N") == 0)
		{
			// Output 3-digit millisecond number zero padded.
			avOut = smvPutNumber3Zero(avOut, vTime.mvMillisecond);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%p") == 0)
		{
			// Output A.M. or P.M. indicator.
			avOut = smvPutAmPm(avOut, vTime.mvHour);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%S") == 0)
		{
			// Output 2-digit second number zero padded.
			avOut = smvPutNumber2Zero(avOut, vTime.mvSecond);
			vFormatOffset += 2;
		}
		else if (vFormat.compare(vFormatOffset, 2, "%t") == 0)
		{
			// Output horizontal tab character.
			avOut = L'\t';
			vFormatOffset += 2;
		}
		else
		{
			// Output normal character.
			avOut = vFormat[vFormatOffset];
			++vFormatOffset;
		}
	}
	return avOut;
}

} /* namespace datetime */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
