//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#ifndef _LFC1_DATETIME_CDATEDURATION_HPP_
#define _LFC1_DATETIME_CDATEDURATION_HPP_

#include <iostream>
#include <string>
#include <mutex>

#include "lfc1/datetime/cfacetbase.hpp"
#include "lfc1/udtsup/cmanipbase.hpp"

namespace lfc1 { namespace datetime {

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This class represents date durations.
///
///	The purpose of this class is to support the increment and decrement operators of the CDate class.
///
class LFC1_DATETIME_API CDateDuration
{
public:
	// Constants
	enum EUnits		///< The list of date durations.
	{
		E_U_YEAR,
		E_U_MONTH,
		E_U_WEEK,
		E_U_DAY
	};

	// Constructors/destructor
	CDateDuration(int avDuration, EUnits avUnit = E_U_DAY) noexcept;
	CDateDuration(const CDateDuration& arRHS) noexcept;
	CDateDuration(CDateDuration&& arRHS) noexcept;
	~CDateDuration() noexcept;

	// Assignment
	CDateDuration& operator=(const CDateDuration& arRHS) noexcept;
	CDateDuration& operator=(CDateDuration&& arRHS) noexcept;

	// Instance setters/getters
	int mvGetDuration() const noexcept;
	EUnits mvGetUnit() const noexcept;

	// Instance special operations
	void swap(CDateDuration& arRHS) noexcept;

private:
	// Instance data
	int mvDuration;
	EUnits mvUnit;
};

LFC1_DATETIME_API void swap(CDateDuration& arLHS, CDateDuration& arRHS) noexcept;

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This class represents a local date.
///
///	Note that this class has no timezone dependency.
///
///	This class supports the following operators:
///	@code
///	binary arithmetic operators: +=, -=
///	unary arithmetic operators:  ++, --
///	@endcode
///
class LFC1_DATETIME_API CDate
{
public:
	// Internal classes
	/// @brief A structure that holds the individual parts of a date.
	///
	struct SDate
	{
		int mvYear;		///< The year.
		int mvMonth;	///< The month.
		int mvDay;		///< The day.
	};

	// Constants
	static const auto	MIN_YEAR		= 0;	///< The minimum year.
	static const auto	MAX_YEAR		= 9999;	///< The maximum year.

	static const auto	MONTH_PER_YEAR	= 12;	///< The number of months in a year.
	static const auto	MIN_MONTH		= 1;	///< The lowest month number in a year.
	static const auto	MAX_MONTH		= 12;	///< The highest month number in a year.

	static const auto	DAY_PER_WEEK	= 7;	///< The number of days per week.
	static const auto	MIN_DAY			= 1;	///< The lowest day number in a month.
	static const auto	MAX_DAY			= 31;	///< The highest day number in a month.

	static const auto	FIRST_DAY_MON	= 0;	///< The number used to indicate that the first day of the
												///< week is Monday.
	static const auto	FIRST_DAY_SUN	= 6;	///< The number used to indicate that the first day of the
												///< week is Sunday.

	static const auto	MIN_DATENO 	= 2298874;	///< Julian day number for 01-Jan-1582.
#if defined(__32__)
	static const auto	MAX_DATENO 	= 2465424;	///< Julian day number for 31-Dec-2037.
#endif
#if defined(__64__)
	static const auto	MAX_DATENO 	= 2817152;	///< Julian day number for 31-Dec-3000.
#endif

	static const auto	DATENO_OFFSET_SUN	= 1;	///< The number to add to a Julian day number such that
													///< when the resulting number is divided by 7, the
													///< remainder is the week day number with Sunday being 0.

	static const auto	MIN_DATE_YEAR	= 1582;	///< The lowest supported year.  This is the year that the
												///< Gregorian calendar started.
#if defined(__32__)
	static const auto	MAX_DATE_YEAR	= 2037;	///< The highest supported year for 32-bit environment.
#endif
#if defined(__64__)
	static const auto	MAX_DATE_YEAR	= 3000;	///< The highest supported year for 64-bit environment.
#endif

	// Class setters/getters
	static int smvGetIndex() noexcept;

	// Class special operations
	static bool smvIsLeapYear(int avYear) noexcept;
	static int smvGetMaxDay(int avYear, int avMonth) noexcept;
	static void smvValidateDate(int avYear, int avMonth, int avDay);

	// Constructors/destructor
	CDate() noexcept;
	CDate(int avDateNo);
	CDate(int avYear, int avMonth, int avDay);
	CDate(const CDate& arRHS) noexcept;
	CDate(CDate&& arRHS) noexcept;
	~CDate() noexcept;

	// Assignment
	CDate& operator=(int avDateNo);
	CDate& operator=(const CDate& arRHS) noexcept;
	CDate& operator=(CDate&& arRHS) noexcept;

	// Conversions
	operator int () const noexcept;

	// Instance setters/getters
	void mvSetDate() noexcept;
	void mvSetDate(int avYear, int avMonth, int avDay);
	SDate mvGetDate() const noexcept;

	// Instance special operations
	void swap(CDate& arRHS) noexcept;
	CDate& operator++();
	CDate operator++(int);
	CDate& operator--();
	CDate operator--(int);
	CDate& operator+=(const CDateDuration& arDateDuration);
	CDate& operator-=(const CDateDuration& arDateDuration);
	std::ios_base::iostate mvInput(std::istream& arStream);
	std::ios_base::iostate mvOutput(std::ostream& arStream) const;

private:
	// Class special operations
	static void smvAllocateIndex() noexcept;

	// Class data
	static int smvIndex;
	static std::once_flag smvFlag;

	// Instance data
	int mvDateNo;
};

LFC1_DATETIME_API void swap(CDate& arLHS, CDate& arRHS) noexcept;

LFC1_DATETIME_API std::istream& operator>>(std::istream& arStream, CDate& arDate);
LFC1_DATETIME_API std::ostream& operator<<(std::ostream& arStream, const CDate& arDate);

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This class serves as the base class for date input/output facets.
///
class LFC1_DATETIME_API CDateBase : public CFacetBase
{
protected:
	// Class special operations
	static std::string smvExpandFormat(std::string avFormat);
	static TInIt smvGetWeekDayName
	(
		TInIt avNext,
		TInIt avEnd,
		std::ios_base& arIosBase,
		std::ios_base::iostate& arIoState,
		int& arWeekDayNo,
		bool avIsShort
	);
	static TInIt smvGetMonthName
	(
		TInIt avNext,
		TInIt avEnd,
		std::ios_base& arIosBase,
		std::ios_base::iostate& arIoState,
		int& arMonthNo,
		bool avIsShort
	);
	static TOutIt smvPutWeekDayName(TOutIt avOut, int avWeekDayNo, bool avIsShort);
	static TOutIt smvPutMonthName(TOutIt avOut, int avMonthNo, bool avIsShort);
	static void smvCalcWeekBasedDetails
	(
		const CDate& arDate,
		int& arWeekBasedYear,
		int* apWeekBasedWeekNo
	) noexcept;
	static int smvCalcWeekNo(const CDate& arDate, int avYear, int avDay) noexcept;

	// Constructors/destructor
	CDateBase() noexcept;
	CDateBase(const CDateBase&) = delete;	///< Not supported.
	CDateBase(CDateBase&&) = delete;	///< Not supported.
	~CDateBase() noexcept override;

	// Assignment
	CDateBase& operator=(const CDateBase&) = delete;	///< Not supported.
	CDateBase& operator=(CDateBase&&) = delete;	///< Not supported.
};

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This class is a CDate class input facet.
///
///	@note This class uses the Template Method design pattern.
///
class LFC1_DATETIME_API CDateGet : public std::locale::facet, public CDateBase
{
public:
	// Class data
	static std::locale::id id;

	// Constructors/destructor
	explicit CDateGet(size_t avRelease = 0);
	CDateGet(const CDateGet&) = delete;	///< Not supported.
	CDateGet(CDateGet&&) = delete;	///< Not supported.
	~CDateGet() noexcept final;

	// Assignment
	CDateGet& operator=(const CDateGet&) = delete;	///< Not supported.
	CDateGet& operator=(CDateGet&&) = delete;	///< Not supported.

	// Instance special operations
	TInIt mvGet
	(
		TInIt avNext,
		TInIt avEnd,
		std::ios_base& arIosBase,
		std::ios_base::iostate& arIoState,
		CDate& arDate,
		const std::string& arFormat
	) const;

protected:
	// Instance special operations
	virtual TInIt mvDoGet
	(
		TInIt avNext,
		TInIt avEnd,
		std::ios_base& arIosBase,
		std::ios_base::iostate& arIoState,
		CDate& arDate,
		const std::string& arFormat
	) const;
};

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This class is a CDate class output facet.
///
///	@note This class uses the Template Method design pattern.
///
class LFC1_DATETIME_API CDatePut : public std::locale::facet, public CDateBase
{
public:
	// Class data
	static std::locale::id id;

	// Constructors/destructor
	explicit CDatePut(size_t avRelease = 0);
	CDatePut(const CDatePut&) = delete;	///< Not supported.
	CDatePut(CDatePut&&) = delete;	///< Not supported.
	~CDatePut() noexcept final;

	// Assignment
	CDatePut& operator=(const CDatePut&) = delete;	///< Not supported.
	CDatePut& operator=(CDatePut&&) = delete;	///< Not supported.

	// Instance special operations
	TOutIt mvPut(TOutIt avOut, const CDate& arDate, const std::string& arFormat) const;

protected:
	// Instance special operations
	virtual TOutIt mvDoPut(TOutIt avOut, const CDate& arDate, const std::string& arFormat) const;
};

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdate.hpp
///	@brief This class is a helper class for the date manipulator.  This class was derived from Sections
///		   3.3.1.3 to 3.3.1.7 of the book Standard C++ IOStreams and Locales.
///
class LFC1_DATETIME_API CDateManip : public lfc1::udtsup::CManipBase<const char*>
{
public:
	// Class special operations
	static std::string smvGetFormat(std::ios_base& arIosBase);
	static std::ios_base::iostate smvCopyFmtErr(std::basic_ios<char>& arIos);

	// Constructors/destructor
	CDateManip(const char* apFormat);
	CDateManip(const CDateManip&) = default;	///< Uses default implementation.
	~CDateManip() noexcept;

	// Assignment
	CDateManip& operator=(const CDateManip&) = default;	///< Uses default implementation.

private:
	// Class special operations
	static void smvSetFormat(std::ios_base& arIosBase, const char* apFormat = "");
	static void smvCallback(std::ios_base::event avEvent, std::ios_base& arIosBase, int avIndex) noexcept;
};

LFC1_DATETIME_API CDateManip setdateformat(const char* apFormat = "");

} /* namespace datetime */ } /* namespace lfc1 */

#endif // _LFC1_DATETIME_CDATEDURATION_HPP_

//============================================================================================================
//	End of Code
//============================================================================================================
