//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#ifndef _LFC1_DATETIME_CDATETIMEDURATION_HPP_
#define _LFC1_DATETIME_CDATETIMEDURATION_HPP_

#include <iostream>
#include <string>
#include <mutex>

#include "lfc1/datetime/cdate.hpp"
#include "lfc1/datetime/ctime.hpp"

namespace lfc1 { namespace datetime {

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdatetime.hpp
///	@brief This class represents date and time durations.
///
///	The purpose of this class is to support the increment and decrement operators of the CDateTime class.
///
class LFC1_DATETIME_API CDateTimeDuration
{
public:
	// Constants
	enum EUnits		/// The list of date and time durations.
	{
		E_U_YEAR,
		E_U_MONTH,
		E_U_WEEK,
		E_U_DAY,
		E_U_HOUR,
		E_U_MINUTE,
		E_U_SECOND,
		E_U_MILLISECOND
	};

	// Constructors/destructor
	CDateTimeDuration(int avDuration, EUnits avUnit = E_U_MILLISECOND) noexcept;
	CDateTimeDuration(const CDateTimeDuration& arRHS) noexcept;
	CDateTimeDuration(CDateTimeDuration&& arRHS) noexcept;
	~CDateTimeDuration() noexcept;

	// Assignment
	CDateTimeDuration& operator=(const CDateTimeDuration& arRHS) noexcept;
	CDateTimeDuration& operator=(CDateTimeDuration&& arRHS) noexcept;

	// Instance setters/getters
	int mvGetDuration() const noexcept;
	EUnits mvGetUnit() const noexcept;

	// Instance special operations
	void swap(CDateTimeDuration& arRHS) noexcept;

private:
	// Instance data
	int mvDuration;
	EUnits mvUnit;
};

LFC1_DATETIME_API void swap(CDateTimeDuration& arLHS, CDateTimeDuration& arRHS) noexcept;

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdatetime.hpp
///	@brief This class represents the time since 01-Jan-1970.
///
///	Note that this class has timezone dependency.
///
///	This class supports the following operators:
///	@code
///	binary arithmetic operators: +=, -=
///	unary arithmetic operators:  ++, --
///	@endcode
///
class LFC1_DATETIME_API CDateTime
{
public:
	// Internal classes
	/// @brief A structure that holds the individual parts of a date and time.
	///
	struct SDateTime
	{
		int mvYear;			///< The year.
		int mvMonth;		///< The month.
		int mvDay;			///< The day.
		int mvHour;			///< The hour.
		int mvMinute;		///< The minute.
		int mvSecond;		///< The second.
		int mvMillisecond;	///< The millisecond.
		int mvDst;			///< The daylight savings indicator.
	};

	// Constants
	static const auto MIN_DATETIMENO = 0LL;					///< DateTimeNo for 01-Jan-1970 00:00:00.000 UTC.
#if defined(__32__)
	static const auto MAX_DATETIMENO = 2145916799999LL;		///< DateTimeNo for 31-Dec-2037 23:59:59.999 UTC.
#endif
#if defined(__64__)
	static const auto MAX_DATETIMENO = 32535215999999LL;	///< DateTimeNo for 31-Dec-3000 23:59:59.999 UTC.
#endif

	// Class setters/getters
	static int smvGetIndex() noexcept;

	// Class special operations.
	static void smvValidateDst(int avDst);

	// Constructors/destructor
	CDateTime() noexcept;
	CDateTime(long long avDateTimeNo);
	CDateTime
	(
		int avYear,
		int avMonth,
		int avDay,
		int avHour,
		int avMinute,
		int avSecond,
		int avMillisecond,
		int avDst = -1
	);
	CDateTime(const CDateTime& arRHS) noexcept;
	CDateTime(CDateTime&& arRHS) noexcept;
	~CDateTime() noexcept;

	// Assignment
	CDateTime& operator=(long long avDateTimeNo);
	CDateTime& operator=(const CDateTime& arRHS) noexcept;
	CDateTime& operator=(CDateTime&& arRHS) noexcept;

	// Conversions
	operator long long () const noexcept;

	// Instance setters/getters
	void mvSetDateTime() noexcept;
	void mvSetDateTime
	(
		int avYear,
		int avMonth,
		int avDay,
		int avHour,
		int avMinute,
		int avSecond,
		int avMillisecond,
		int avDst = -1
	);
	SDateTime mvGetDateTime() const noexcept;

	// Instance special operations
	void swap(CDateTime& arRHS) noexcept;
	CDateTime& operator++();
	CDateTime operator++(int);
	CDateTime& operator--();
	CDateTime operator--(int);
	CDateTime& operator+=(const CDateTimeDuration& arDateTimeDuration);
	CDateTime& operator-=(const CDateTimeDuration& arDateTimeDuration);
	std::ios_base::iostate mvInput(std::istream& arStream);
	std::ios_base::iostate mvOutput(std::ostream& arStream) const;

private:
	// Class special operations
	static void smvAllocateIndex() noexcept;

	// Class data
	static int smvIndex;
	static std::once_flag smvFlag;

	// Instance data
	long long mvDateTimeNo;
};

LFC1_DATETIME_API void swap(CDateTime& arLHS, CDateTime& arRHS) noexcept;

LFC1_DATETIME_API std::istream& operator>>(std::istream& arStream, CDateTime& arDateTime);
LFC1_DATETIME_API std::ostream& operator<<(std::ostream& arStream, const CDateTime& arDateTime);

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdatetime.hpp
///	@brief This class serves as the base class for date and time input/output facets.
///
class LFC1_DATETIME_API CDateTimeBase : public CDateBase, public CTimeBase
{
protected:
	// Class special operations
	static std::string smvExpandFormat(std::string avFormat);
	static TInIt smvGetTzName(TInIt arNext, TInIt arEnd, std::ios_base::iostate& arIoState, int& arDst);
	static TInIt smvGetTzOffset
	(
		TInIt arNext,
		TInIt arEnd,
		std::ios_base& arIosBase,
		std::ios_base::iostate& arIoState,
		int& arDst
	);
	static TOutIt smvPutTzName(TOutIt avOut, int avDst);
	static TOutIt smvPutTzOffset(TOutIt avOut, int avDst) noexcept;

	// Constructors/destructor
	CDateTimeBase() noexcept;
	CDateTimeBase(const CDateTimeBase&) = delete;	///< Not supported.
	CDateTimeBase(CDateTimeBase&&) = delete;	///< Not supported.
	~CDateTimeBase() noexcept override;

	// Assignment
	CDateTimeBase& operator=(const CDateTimeBase&) = delete;	///< Not supported.
	CDateTimeBase& operator=(CDateTimeBase&&) = delete;	///< Not supported.
};

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdatetime.hpp
///	@brief This class is a CDateTime class input facet.
///
///	@note This class uses the Template Method design pattern.
///
class LFC1_DATETIME_API CDateTimeGet : public std::locale::facet, public CDateTimeBase
{
public:
	// Class data
	static std::locale::id id;

	// Constructors/destructor
	explicit CDateTimeGet(size_t avRelease = 0);
	CDateTimeGet(const CDateTimeGet&) = delete;	///< Not supported.
	CDateTimeGet(CDateTimeGet&&) = delete;	///< Not supported.
	~CDateTimeGet() noexcept final;

	// Assignment
	CDateTimeGet& operator=(const CDateTimeGet&) = delete;	///< Not supported.
	CDateTimeGet& operator=(CDateTimeGet&&) = delete;	///< Not supported.

	// Instance special operations
	TInIt mvGet
	(
		TInIt avNext,
		TInIt avEnd,
		std::ios_base& arIosBase,
		std::ios_base::iostate& arIoState,
		CDateTime& arDateTime,
		const std::string& arFormat
	) const;

protected:
	// Instance special operations
	virtual TInIt mvDoGet
	(
		TInIt avNext,
		TInIt avEnd,
		std::ios_base& arIosBase,
		std::ios_base::iostate& arIoState,
		CDateTime& arDateTime,
		const std::string& arFormat
	) const;
};

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdatetime.hpp
///	@brief This class is a CDateTime class output facet.
///
///	@note This class uses the Template Method design pattern.
///
class LFC1_DATETIME_API CDateTimePut : public std::locale::facet, public CDateTimeBase
{
public:
	// Class data
	static std::locale::id id;

	// Constructors/destructor
	explicit CDateTimePut(size_t avRelease = 0);
	CDateTimePut(const CDateTimePut&) = delete;	///< Not supported.
	CDateTimePut(CDateTimePut&&) = delete;	///< Not supported.
	~CDateTimePut() noexcept final;

	// Assignment
	CDateTimePut& operator=(const CDateTimePut&) = delete;	///< Not supported.
	CDateTimePut& operator=(CDateTimePut&&) = delete;	///< Not supported.

	// Instance special operations
	TOutIt mvPut(TOutIt avOut, const CDateTime& arDateTime, const std::string& arFormat) const;

protected:
	// Instance special operations
	virtual TOutIt mvDoPut(TOutIt avOut, const CDateTime& arDateTime, const std::string& arFormat) const;
};

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/cdatetime.hpp
///	@brief This class is a helper class for the date and time manipulator.  This class was derived from
///		   Section 3.3.1.3 to 3.3.1.7 of the book Standard C++ IOStreams and Locales.
///
class LFC1_DATETIME_API CDateTimeManip : public lfc1::udtsup::CManipBase<const char*>
{
public:
	// Class special operations
	static std::string smvGetFormat(std::ios_base& arIosBase);
	static std::ios_base::iostate smvCopyFmtErr(std::basic_ios<char>& arIos);

	// Constructors/destructor
	CDateTimeManip(const char* apFormat);
	CDateTimeManip(const CDateTimeManip&) = default;	///< Uses default implementation.
	~CDateTimeManip() noexcept;

	// Assignment
	CDateTimeManip& operator=(const CDateTimeManip&) = default;	///< Uses default implementation.

private:
	// Class special operations
	static void smvSetFormat(std::ios_base& arIosBase, const char* apFormat = "");
	static void smvCallback(std::ios_base::event avEvent, std::ios_base& arIosBase, int avIndex) noexcept;
};

LFC1_DATETIME_API CDateTimeManip setdatetimeformat(const char* apFormat = "");

} /* namespace datetime */ } /* namespace lfc1 */

#endif // _LFC1_DATETIME_CDATETIMEDURATION_HPP_

//============================================================================================================
//	End of Code
//============================================================================================================
