//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#ifndef _LFC1_DATETIME_CTIMEDURATION_HPP_
#define _LFC1_DATETIME_CTIMEDURATION_HPP_

#include <iostream>
#include <string>
#include <mutex>

#include "lfc1/datetime/cfacetbase.hpp"
#include "lfc1/udtsup/cmanipbase.hpp"

namespace lfc1 { namespace datetime {

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/ctime.hpp
///	@brief This class represents time durations.
///
///	The purpose of this class is to support the increment and decrement operators of the CTime class.
///
class LFC1_DATETIME_API CTimeDuration
{
public:
	// Constants
	enum EUnits		///< The list of time durations.
	{
		E_U_HOUR,
		E_U_MINUTE,
		E_U_SECOND,
		E_U_MILLISECOND
	};

	// Constructors/destructor
	CTimeDuration(int avDuration, EUnits avUnit = E_U_MILLISECOND) noexcept;
	CTimeDuration(const CTimeDuration& arRHS) noexcept;
	CTimeDuration(CTimeDuration&& arRHS) noexcept;
	~CTimeDuration() noexcept;

	// Assignment
	CTimeDuration& operator=(const CTimeDuration& arRHS) noexcept;
	CTimeDuration& operator=(CTimeDuration&& arRHS) noexcept;

	// Instance setters/getters
	int mvGetDuration() const noexcept;
	EUnits mvGetUnit() const noexcept;

	// Instance special operations
	void swap(CTimeDuration& arRHS) noexcept;

private:
	// Instance data
	int mvDuration;
	EUnits mvUnit;
};

LFC1_DATETIME_API void swap(CTimeDuration& arLHS, CTimeDuration& arRHS) noexcept;

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/ctime.hpp
///	@brief This class represents the time since midnight.
///
///	This class supports the following operators:
///	@code
///	binary arithmetic operators: +=, -=
///	unary arithmetic operators:  ++, --
///	@endcode
///
class LFC1_DATETIME_API CTime
{
public:
	// Internal classes
	/// @brief A structure that holds the individual parts of a time.
	///
	struct STime
	{
		int mvHour;			///< The hour.
		int mvMinute;		///< The minute.
		int mvSecond;		///< The second.
		int mvMillisecond;	///< The millisecond.
	};

	// Constants
	static const auto HR_PER_DAY		= 24;				///< The number of hours in a day.
	static const auto MIN_HR			= 0;				///< The lowest hour number in a day.
	static constexpr auto MAX_HR		= HR_PER_DAY - 1;	///< The highest hour number in a day.

	static const auto MIN_PER_HR		= 60;				///< The number of minutes in an hour.
	static const auto MIN_MIN			= 0;				///< The lowest minute number in an hour.
	static constexpr auto MAX_MIN		= MIN_PER_HR - 1;	///< The highest minute number in an hour.

	static const auto SEC_PER_MIN		= 60;				///< The number of seconds in a minute.
	static const auto MIN_SEC			= 0;				///< The lowest second number in a minute.
	static constexpr auto MAX_SEC		= SEC_PER_MIN - 1;	///< The highest second number in a minute.

	static const auto MSEC_PER_SEC		= 1000;				///< The number of milliseconds in a second.
	static const auto MIN_MSEC			= 0;				///< The lowest millisecond number in a second.
	static constexpr auto MAX_MSEC		= MSEC_PER_SEC - 1;	///< The highest millisecond number in a second.

	static constexpr auto MSEC_PER_MIN	= SEC_PER_MIN * MSEC_PER_SEC;	///< The no. of millisecs. per minute.
	static constexpr auto MSEC_PER_HR	= MIN_PER_HR * MSEC_PER_MIN;	///< The no. of milliseconds per hour.
	static constexpr auto MSEC_PER_DAY	= HR_PER_DAY * MSEC_PER_HR;		///< The no. of milliseconds per day.

	static constexpr auto SEC_PER_HR	= MIN_PER_HR * SEC_PER_MIN;	///< The number of seconds per hour.
	static constexpr auto SEC_PER_DAY	= HR_PER_DAY * SEC_PER_HR;	///< The number of seconds per day.

	static constexpr auto MIN_PER_DAY	= HR_PER_DAY * MIN_PER_HR;	///< The number of minutes per day.

	static const auto MIN_TIMENO		= 0;				///< The lowest millisecond of a day.
	static constexpr auto MAX_TIMENO	= MSEC_PER_DAY - 1;	///< The highest millisecond of a day.

	// Class setters/getters
	static int smvGetIndex() noexcept;

	// Class special operations
	static void smvValidateTime(int avHour, int avMinute, int avSecond, int avMillisecond);

	// Constructors/destructor
	CTime() noexcept;
	CTime(int avTimeNo);
	CTime(int avHour, int avMinute, int avSecond, int avMillisecond);
	CTime(const CTime& arRHS) noexcept;
	CTime(CTime&& arRHS) noexcept;
	~CTime() noexcept;

	// Assignment
	CTime& operator=(int avTimeNo);
	CTime& operator=(const CTime& arRHS) noexcept;
	CTime& operator=(CTime&& arRHS) noexcept;

	// Conversions
	operator int () const noexcept;

	// Instance setters/getters
	void mvSetTime() noexcept;
	void mvSetTime(int avHour, int avMinute, int avSecond, int avMillisecond);
	STime mvGetTime() const noexcept;

	// Instance special operations
	void swap(CTime& arRHS) noexcept;
	CTime& operator++();
	CTime operator++(int);
	CTime& operator--();
	CTime operator--(int);
	CTime& operator+=(const CTimeDuration& arTimeDuration);
	CTime& operator-=(const CTimeDuration& arTimeDuration);
	std::ios_base::iostate mvInput(std::istream& arStream);
	std::ios_base::iostate mvOutput(std::ostream& arStream) const;

private:
	// Class special operations
	static void smvAllocateIndex() noexcept;

	// Class data
	static int smvIndex;
	static std::once_flag smvFlag;

	// Instance data
	int mvTimeNo;
};

LFC1_DATETIME_API void swap(CTime& arLHS, CTime& arRHS) noexcept;

LFC1_DATETIME_API std::istream& operator>>(std::istream& arStream, CTime& arTime);
LFC1_DATETIME_API std::ostream& operator<<(std::ostream& arStream, const CTime& arTime);

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/ctime.hpp
///	@brief This class serves as the base class for time input/output facets.
///
class LFC1_DATETIME_API CTimeBase : public CFacetBase
{
protected:
	// Class special operations
	static std::string smvExpandFormat(std::string avFormat);
	static TInIt smvGetAmPm
	(
		TInIt avNext,
		TInIt avEnd,
		std::ios_base& arIosBase,
		std::ios_base::iostate& arIoState,
		bool& arHasAmPm,
		bool& arIsPm
	);
	static TOutIt smvPutAmPm(TOutIt avOut, int avHour);

	// Constructors/destructor
	CTimeBase() noexcept;
	CTimeBase(const CTimeBase&) = delete;	///< Not supported.
	CTimeBase(CTimeBase&&) = delete;	///< Not supported.
	~CTimeBase() noexcept override;

	// Assignment
	CTimeBase& operator=(const CTimeBase&) = delete;	///< Not supported.
	CTimeBase& operator=(CTimeBase&&) = delete;	///< Not supported.
};

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/ctime.hpp
///	@brief This class is a CTime class input facet.
///
///	@note This class uses the Template Method design pattern.
///
class LFC1_DATETIME_API CTimeGet : public std::locale::facet, public CTimeBase
{
public:
	// Class data
	static std::locale::id id;

	// Constructors/destructor
	explicit CTimeGet(size_t avRelease = 0);
	CTimeGet(const CTimeGet&) = delete;	///< Not supported.
	CTimeGet(CTimeGet&&) = delete;	///< Not supported.
	~CTimeGet() noexcept final;

	// Assignment
	CTimeGet& operator=(const CTimeGet&) = delete;	///< Not supported.
	CTimeGet& operator=(CTimeGet&&) = delete;	///< Not supported.

	// Instance special operations
	TInIt mvGet
	(
		TInIt avNext,
		TInIt avEnd,
		std::ios_base& arIosBase,
		std::ios_base::iostate& arIoState,
		CTime& arTime,
		const std::string& arFormat
	) const;

protected:
	// Instance special operations
	virtual TInIt mvDoGet
	(
		TInIt avNext,
		TInIt avEnd,
		std::ios_base& arIosBase,
		std::ios_base::iostate& arIoState,
		CTime& arTime,
		const std::string& arFormat
	) const;
};

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/ctime.hpp
///	@brief This class is a CTime class output facet.
///
///	@note This class uses the Template Method design pattern.
///
class LFC1_DATETIME_API CTimePut : public std::locale::facet, public CTimeBase
{
public:
	// Class data
	static std::locale::id id;

	// Constructors/destructor
	explicit CTimePut(size_t avRelease = 0);
	CTimePut(const CTimePut&) = delete;	///< Not supported.
	CTimePut(CTimePut&&) = delete;	///< Not supported.
	~CTimePut() noexcept final;

	// Assignment
	CTimePut& operator=(const CTimePut&) = delete;	///< Not supported.
	CTimePut& operator=(CTimePut&&) = delete;	///< Not supported.

	// Instance special operations
	TOutIt mvPut(TOutIt avOut, const CTime& arTime, const std::string& arFormat) const;

protected:
	// Instance special operations
	virtual TOutIt mvDoPut(TOutIt avOut, const CTime& arTime, const std::string& arFormat) const;
};

///	@ingroup LFC1_DATETIME
///	@headerfile lfc1/datetime/ctime.hpp
///	@brief This class is a helper class for the time manipulator.  This class was derived from Sections
///		   3.3.1.3 to 3.3.1.7 of the book Standard C++ IOStreams and Locales.
///
class LFC1_DATETIME_API CTimeManip : public lfc1::udtsup::CManipBase<const char*>
{
public:
	// Class special operations
	static std::string smvGetFormat(std::ios_base& arIosBase);
	static std::ios_base::iostate smvCopyFmtErr(std::basic_ios<char>& arIos);

	// Constructors/destructor
	CTimeManip(const char* apFormat);
	CTimeManip(const CTimeManip&) = default;	///< Uses default implementation.
	~CTimeManip() noexcept;

	// Assignment
	CTimeManip& operator=(const CTimeManip&) = default;	///< Uses default implementation.

private:
	// Class special operations
	static void smvSetFormat(std::ios_base& arIosBase, const char* apFormat = "");
	static void smvCallback(std::ios_base::event avEvent, std::ios_base& arIosBase, int avIndex) noexcept;
};

LFC1_DATETIME_API CTimeManip settimeformat(const char* apFormat = "");

} /* namespace datetime */ } /* namespace lfc1 */

#endif // _LFC1_DATETIME_CTIMEDURATION_HPP_

//============================================================================================================
//	End of Code
//============================================================================================================
